# Copyright 2013 Canonical Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Ceph Backup Service Implementation.

This driver supports backing up volumes of any type to a Ceph object store. It
is also capable of detecting whether the volume to be backed up is a Ceph RBD
volume and, if so, attempts to perform incremental/differential backups.

Support is also included for the following in the case of a source volume being
a Ceph RBD volume:

    * backing up within the same Ceph pool (not recommended)
    * backing up between different Ceph pools
    * backing up between different Ceph clusters

At the time of writing, differential backup support in Ceph/librbd was quite
new so this driver accounts for this by first attempting differential backup
and falling back to full backup/copy if the former fails. It is recommended
that you upgrade to Ceph Dumpling (>= v0.67) or above to get the best results.

If incremental backups are used, multiple backups of the same volume are stored
as snapshots so that minimal space is consumed in the object store and
restoring the volume takes a far reduced amount of time compared to a full
copy.

Note that Cinder supports restoring to a new volume or the original volume the
backup was taken from. For the latter case, a full copy is enforced since this
was deemed the safest action to take. It is therefore recommended to always
restore to a new volume (default).
"""

import fcntl
import json
import os
import re
import subprocess
import time

import eventlet
from os_brick.initiator import linuxrbd
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import units
from six.moves import range

from cinder.backup import driver
from cinder import exception
from cinder.i18n import _
from cinder import interface
from cinder import utils
import cinder.volume.drivers.rbd as rbd_driver

try:
    import rados
    import rbd
except ImportError:
    rados = None
    rbd = None

LOG = logging.getLogger(__name__)

service_opts = [
    cfg.StrOpt('backup_ceph_conf', default='/etc/ceph/ceph.conf',
               help='Ceph configuration file to use.'),
    cfg.StrOpt('backup_ceph_user', default='cinder',
               help='The Ceph user to connect with. Default here is to use '
                    'the same user as for Cinder volumes. If not using cephx '
                    'this should be set to None.'),
    cfg.IntOpt('backup_ceph_chunk_size', default=(units.Mi * 128),
               help='The chunk size, in bytes, that a backup is broken into '
                    'before transfer to the Ceph object store.'),
    cfg.StrOpt('backup_ceph_pool', default='backups',
               help='The Ceph pool where volume backups are stored.'),
    cfg.IntOpt('backup_ceph_stripe_unit', default=0,
               help='RBD stripe unit to use when creating a backup image.'),
    cfg.IntOpt('backup_ceph_stripe_count', default=0,
               help='RBD stripe count to use when creating a backup image.'),
    cfg.BoolOpt('backup_ceph_image_journals', default=False,
                help='If True, apply JOURNALING and EXCLUSIVE_LOCK feature '
                     'bits to the backup RBD objects to allow mirroring'),
    cfg.BoolOpt('restore_discard_excess_bytes', default=True,
                help='If True, always discard excess bytes when restoring '
                     'volumes i.e. pad with zeroes.')
]

CONF = cfg.CONF
CONF.register_opts(service_opts)


class VolumeMetadataBackup(object):

    def __init__(self, client, backup_id):
        self._client = client
        self._backup_id = backup_id

    @property
    def name(self):
        return utils.convert_str("backup.%s.meta" % self._backup_id)

    @property
    def exists(self):
        meta_obj = eventlet.tpool.Proxy(rados.Object(self._client.ioctx,
                                                     self.name))
        return self._exists(meta_obj)

    def _exists(self, obj):
        try:
            obj.stat()
        except rados.ObjectNotFound:
            return False
        else:
            return True

    def set(self, json_meta):
        """Write JSON metadata to a new object.

        This should only be called once per backup. Raises
        VolumeMetadataBackupExists if the object already exists.
        """
        meta_obj = eventlet.tpool.Proxy(rados.Object(self._client.ioctx,
                                                     self.name))
        if self._exists(meta_obj):
            msg = _("Metadata backup object '%s' already exists") % self.name
            raise exception.VolumeMetadataBackupExists(msg)

        meta_obj.write(json_meta.encode('utf-8'))

    def get(self):
        """Get metadata backup object.

        Returns None if the object does not exist.
        """
        meta_obj = eventlet.tpool.Proxy(rados.Object(self._client.ioctx,
                                                     self.name))
        if not self._exists(meta_obj):
            LOG.debug("Metadata backup object %s does not exist", self.name)
            return None

        return meta_obj.read().decode('utf-8')

    def remove_if_exists(self):
        meta_obj = eventlet.tpool.Proxy(rados.Object(self._client.ioctx,
                                                     self.name))
        try:
            meta_obj.remove()
        except rados.ObjectNotFound:
            LOG.debug("Metadata backup object '%s' not found - ignoring",
                      self.name)


@interface.backupdriver
class CephBackupDriver(driver.BackupDriver):
    """Backup Cinder volumes to Ceph Object Store.

    This class enables backing up Cinder volumes to a Ceph object store.
    Backups may be stored in their own pool or even cluster. Store location is
    defined by the Ceph conf file and service config options supplied.

    If the source volume is itself an RBD volume, the backup will be performed
    using incremental differential backups which *should* give a performance
    gain.
    """

    def __init__(self, context, db=None, execute=None):
        super(CephBackupDriver, self).__init__(context, db)
        self.rbd = rbd
        self.rados = rados
        self.chunk_size = CONF.backup_ceph_chunk_size
        self._execute = execute or utils.execute

        if self._supports_stripingv2:
            self.rbd_stripe_unit = CONF.backup_ceph_stripe_unit
            self.rbd_stripe_count = CONF.backup_ceph_stripe_count
        else:
            LOG.info("RBD striping not supported - ignoring configuration "
                     "settings for rbd striping.")
            self.rbd_stripe_count = 0
            self.rbd_stripe_unit = 0

        self._ceph_backup_user = utils.convert_str(CONF.backup_ceph_user)
        self._ceph_backup_pool = utils.convert_str(CONF.backup_ceph_pool)
        self._ceph_backup_conf = utils.convert_str(CONF.backup_ceph_conf)

    @staticmethod
    def get_driver_options():
        return service_opts

    def _validate_string_args(self, *args):
        """Ensure all args are non-None and non-empty."""
        return all(args)

    def _ceph_args(self, user, conf=None, pool=None):
        """Create default ceph args for executing rbd commands.

        If no --conf is provided, rbd will look in the default locations e.g.
        /etc/ceph/ceph.conf
        """

        # Make sure user arg is valid since rbd command may not fail if
        # invalid/no user provided, resulting in unexpected behaviour.
        if not self._validate_string_args(user):
            raise exception.BackupInvalidCephArgs(_("invalid user '%s'") %
                                                  user)

        args = ['--id', user]
        if conf:
            args.extend(['--conf', conf])
        if pool:
            args.extend(['--pool', pool])

        return args

    @property
    def _supports_layering(self):
        """Determine if copy-on-write is supported by our version of librbd."""
        return hasattr(self.rbd, 'RBD_FEATURE_LAYERING')

    @property
    def _supports_stripingv2(self):
        """Determine if striping is supported by our version of librbd."""
        return hasattr(self.rbd, 'RBD_FEATURE_STRIPINGV2')

    @property
    def _supports_exclusive_lock(self):
        """Determine if exclusive-lock is supported by librbd."""
        return hasattr(self.rbd, 'RBD_FEATURE_EXCLUSIVE_LOCK')

    @property
    def _supports_journaling(self):
        """Determine if journaling is supported by our version of librbd."""
        return hasattr(self.rbd, 'RBD_FEATURE_JOURNALING')

    def _get_rbd_support(self):
        """Determine RBD features supported by our version of librbd."""
        old_format = True
        features = 0
        if self._supports_layering:
            old_format = False
            features |= self.rbd.RBD_FEATURE_LAYERING
        if self._supports_stripingv2:
            old_format = False
            features |= self.rbd.RBD_FEATURE_STRIPINGV2

        # journaling requires exclusive_lock; check both together
        if CONF.backup_ceph_image_journals:
            if self._supports_exclusive_lock and self._supports_journaling:
                old_format = False
                features |= (self.rbd.RBD_FEATURE_EXCLUSIVE_LOCK |
                             self.rbd.RBD_FEATURE_JOURNALING)
            else:
                # FIXME (tasker): when the backup manager supports loading the
                #   driver during its initialization, this exception should be
                #   moved to the driver's initialization so that it can stop
                #   the service from starting when the underyling RBD does not
                #   support the requested features.
                LOG.error("RBD journaling not supported - unable to "
                          "support per image mirroring in backup pool")
                raise exception.BackupInvalidCephArgs(
                    _("Image Journaling set but RBD backend does "
                      "not support journaling")
                )

        return (old_format, features)

    def check_for_setup_error(self):
        """Returns an error if prerequisites aren't met."""
        if rados is None or rbd is None:
            msg = _('rados and rbd python libraries not found')
            raise exception.BackupDriverException(reason=msg)

        for attr in ['backup_ceph_user', 'backup_ceph_pool',
                     'backup_ceph_conf']:
            val = getattr(CONF, attr)
            if not val:
                raise exception.InvalidConfigurationValue(option=attr,
                                                          value=val)
        # NOTE: Checking connection to ceph
        # RADOSClient __init__ method invokes _connect_to_rados
        # so no need to check for self.rados.Error here.
        with rbd_driver.RADOSClient(self, self._ceph_backup_pool):
            pass

    def _connect_to_rados(self, pool=None):
        """Establish connection to the backup Ceph cluster."""
        client = eventlet.tpool.Proxy(self.rados.Rados(
                                      rados_id=self._ceph_backup_user,
                                      conffile=self._ceph_backup_conf))
        try:
            client.connect()
            pool_to_open = utils.convert_str(pool or self._ceph_backup_pool)
            ioctx = client.open_ioctx(pool_to_open)
            return client, ioctx
        except self.rados.Error:
            # shutdown cannot raise an exception
            client.shutdown()
            raise

    def _disconnect_from_rados(self, client, ioctx):
        """Terminate connection with the backup Ceph cluster."""
        # closing an ioctx cannot raise an exception
        ioctx.close()
        client.shutdown()

    def _format_base_name(self, service_metadata):
        base_name = json.loads(service_metadata)["base"]
        return utils.convert_str(base_name)

    def _get_backup_base_name(self, volume_id, backup=None):
        """Return name of base image used for backup.

        Incremental backups use a new base name so we support old and new style
        format.
        """
        # Ensure no unicode
        if not backup:
            return utils.convert_str("volume-%s.backup.base" % volume_id)

        if backup.service_metadata:
            return self._format_base_name(backup.service_metadata)

        # 'parent' field will only be present in incremental backups. This is
        # filled by cinder-api
        if backup.parent:
            # Old backups don't have the base name in the service_metadata,
            # so we use the default RBD backup base
            if backup.parent.service_metadata:
                service_metadata = backup.parent.service_metadata
                base_name = self._format_base_name(service_metadata)
            else:
                base_name = utils.convert_str("volume-%s.backup.base"
                                              % volume_id)

            return base_name

        return utils.convert_str("volume-%s.backup.%s"
                                 % (volume_id, backup.id))

    def _discard_bytes(self, volume, offset, length):
        """Trim length bytes from offset.

        If the volume is an rbd do a discard() otherwise assume it is a file
        and pad with zeroes.
        """
        if length:
            LOG.debug("Discarding %(length)s bytes from offset %(offset)s",
                      {'length': length, 'offset': offset})
            if self._file_is_rbd(volume):
                limit = 2 * units.Gi - 1
                chunks = int(length / limit)
                for chunk in range(0, chunks):
                    eventlet.tpool.Proxy(volume.rbd_image).discard(
                        offset + chunk * limit, limit)
                rem = int(length % limit)
                if rem:
                    eventlet.tpool.Proxy(volume.rbd_image).discard(
                        offset + chunks * limit, rem)
            else:
                zeroes = '\0' * self.chunk_size
                chunks = int(length / self.chunk_size)
                for chunk in range(0, chunks):
                    LOG.debug("Writing zeroes chunk %d", chunk)
                    volume.write(zeroes)
                    volume.flush()

                rem = int(length % self.chunk_size)
                if rem:
                    zeroes = '\0' * rem
                    volume.write(zeroes)
                    volume.flush()

    def _transfer_data(self, src, src_name, dest, dest_name, length):
        """Transfer data between files (Python IO objects)."""
        LOG.debug("Transferring data between '%(src)s' and '%(dest)s'",
                  {'src': src_name, 'dest': dest_name})

        chunks = int(length / self.chunk_size)
        LOG.debug("%(chunks)s chunks of %(bytes)s bytes to be transferred",
                  {'chunks': chunks, 'bytes': self.chunk_size})

        for chunk in range(0, chunks):
            before = time.time()
            data = src.read(self.chunk_size)
            # If we have reach end of source, discard any extraneous bytes from
            # destination volume if trim is enabled and stop writing.
            if data == b'':
                if CONF.restore_discard_excess_bytes:
                    self._discard_bytes(dest, dest.tell(),
                                        length - dest.tell())

                return

            dest.write(data)
            dest.flush()
            delta = (time.time() - before)
            rate = (self.chunk_size / delta) / 1024
            LOG.debug("Transferred chunk %(chunk)s of %(chunks)s "
                      "(%(rate)dK/s)",
                      {'chunk': chunk + 1,
                       'chunks': chunks,
                       'rate': rate})

        rem = int(length % self.chunk_size)
        if rem:
            LOG.debug("Transferring remaining %s bytes", rem)
            data = src.read(rem)
            if data == b'':
                if CONF.restore_discard_excess_bytes:
                    self._discard_bytes(dest, dest.tell(), rem)
            else:
                dest.write(data)
                dest.flush()

    def _create_base_image(self, name, size, rados_client):
        """Create a base backup image.

        This will be the base image used for storing differential exports.
        """
        LOG.debug("Creating base image '%s'", name)
        old_format, features = self._get_rbd_support()
        eventlet.tpool.Proxy(self.rbd.RBD()).create(
            ioctx=rados_client.ioctx,
            name=name,
            size=size,
            old_format=old_format,
            features=features,
            stripe_unit=self.rbd_stripe_unit,
            stripe_count=self.rbd_stripe_count)

    def _delete_backup_snapshot(self, rados_client, base_name, backup_id):
        """Delete snapshot associated with this backup if one exists.

        A backup should have at most ONE associated snapshot.

        This is required before attempting to delete the base image. The
        snapshot on the original volume can be left as it will be purged when
        the volume is deleted.

        Returns tuple(deleted_snap_name, num_of_remaining_snaps).
        """
        remaining_snaps = 0
        base_rbd = eventlet.tpool.Proxy(self.rbd.Image(rados_client.ioctx,
                                                       base_name))
        try:
            snap_name = self._get_backup_snap_name(base_rbd, base_name,
                                                   backup_id)
            if snap_name:
                LOG.debug("Deleting backup snapshot='%s'", snap_name)
                base_rbd.remove_snap(snap_name)
            else:
                LOG.debug("No backup snapshot to delete")

            # Now check whether any snapshots remain on the base image
            backup_snaps = self.get_backup_snaps(base_rbd)
            if backup_snaps:
                remaining_snaps = len(backup_snaps)
        finally:
            base_rbd.close()

        return snap_name, remaining_snaps

    def _try_delete_base_image(self, backup, base_name=None):
        """Try to delete backup RBD image.

        If the rbd image is a base image for incremental backups, it may have
        snapshots. Delete the snapshot associated with backup_id and if the
        image has no more snapshots, delete it. Otherwise return.

        If no base name is provided try normal (full) format then diff format
        image name.

        If a base name is provided but does not exist, ImageNotFound will be
        raised.

        If the image is busy, a number of retries will be performed if
        ImageBusy is received, after which the exception will be propagated to
        the caller.
        """
        retries = 3
        delay = 5
        try_diff_format = False
        volume_id = backup.volume_id

        if base_name is None:
            try_diff_format = True

            base_name = self._get_backup_base_name(volume_id, backup=backup)
            LOG.debug("Trying diff format basename='%(basename)s' for "
                      "backup base image of volume %(volume)s.",
                      {'basename': base_name, 'volume': volume_id})

        with eventlet.tpool.Proxy(rbd_driver.RADOSClient(self,
                                  backup.container)) as client:
            rbd_exists, base_name = \
                self._rbd_image_exists(base_name, volume_id, client,
                                       try_diff_format=try_diff_format)
            if not rbd_exists:
                raise self.rbd.ImageNotFound(_("image %s not found") %
                                             base_name)

            while retries >= 0:
                # First delete associated snapshot from base image (if exists)
                snap, rem = self._delete_backup_snapshot(client, base_name,
                                                         backup.id)
                if rem:
                    LOG.info(
                        "Backup base image of volume %(volume)s still "
                        "has %(snapshots)s snapshots so skipping base "
                        "image delete.",
                        {'snapshots': rem, 'volume': volume_id})
                    return

                LOG.info("Deleting backup base image='%(basename)s' of "
                         "volume %(volume)s.",
                         {'basename': base_name, 'volume': volume_id})
                # Delete base if no more snapshots
                try:
                    eventlet.tpool.Proxy(self.rbd.RBD()).remove(
                        client.ioctx, base_name)
                except self.rbd.ImageBusy:
                    # Allow a retry if the image is busy
                    if retries > 0:
                        LOG.info("Backup image of volume %(volume)s is "
                                 "busy, retrying %(retries)s more time(s) "
                                 "in %(delay)ss.",
                                 {'retries': retries,
                                  'delay': delay,
                                  'volume': volume_id})
                    else:
                        LOG.error("Max retries reached deleting backup "
                                  "%(basename)s image of volume %(volume)s.",
                                  {'volume': volume_id,
                                   'basename': base_name})
                        raise
                else:
                    LOG.debug("Base backup image='%(basename)s' of volume "
                              "%(volume)s deleted.",
                              {'basename': base_name, 'volume': volume_id})
                    retries = 0
                finally:
                    retries -= 1

            # Since we have deleted the base image we can delete the source
            # volume backup snapshot.
            src_name = utils.convert_str(volume_id)
            if src_name in eventlet.tpool.Proxy(
                    self.rbd.RBD()).list(client.ioctx):
                LOG.debug("Deleting source volume snapshot '%(snapshot)s' "
                          "for backup %(basename)s.",
                          {'snapshot': snap, 'basename': base_name})
                src_rbd = eventlet.tpool.Proxy(self.rbd.Image(client.ioctx,
                                                              src_name))
                try:
                    src_rbd.remove_snap(snap)
                finally:
                    src_rbd.close()

    def _piped_execute(self, cmd1, cmd2):
        """Pipe output of cmd1 into cmd2."""
        LOG.debug("Piping cmd1='%s' into...", ' '.join(cmd1))
        LOG.debug("cmd2='%s'", ' '.join(cmd2))

        try:
            p1 = subprocess.Popen(cmd1, stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE,
                                  close_fds=True)
        except OSError as e:
            LOG.error("Pipe1 failed - %s ", e)
            raise

        # NOTE(dosaboy): ensure that the pipe is blocking. This is to work
        # around the case where evenlet.green.subprocess is used which seems to
        # use a non-blocking pipe.
        flags = fcntl.fcntl(p1.stdout, fcntl.F_GETFL) & (~os.O_NONBLOCK)
        fcntl.fcntl(p1.stdout, fcntl.F_SETFL, flags)

        try:
            p2 = subprocess.Popen(cmd2, stdin=p1.stdout,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE,
                                  close_fds=True)
        except OSError as e:
            LOG.error("Pipe2 failed - %s ", e)
            raise

        p1.stdout.close()
        stdout, stderr = p2.communicate()
        return p2.returncode, stderr

    def _rbd_diff_transfer(self, src_name, src_pool, dest_name, dest_pool,
                           src_user, src_conf, dest_user, dest_conf,
                           src_snap=None, from_snap=None):
        """Copy only extents changed between two points.

        If no snapshot is provided, the diff extents will be all those changed
        since the rbd volume/base was created, otherwise it will be those
        changed since the snapshot was created.
        """
        LOG.debug("Performing differential transfer from '%(src)s' to "
                  "'%(dest)s'",
                  {'src': src_name, 'dest': dest_name})

        # NOTE(dosaboy): Need to be tolerant of clusters/clients that do
        # not support these operations since at the time of writing they
        # were very new.

        src_ceph_args = self._ceph_args(src_user, src_conf, pool=src_pool)
        dest_ceph_args = self._ceph_args(dest_user, dest_conf, pool=dest_pool)

        cmd1 = ['rbd', 'export-diff'] + src_ceph_args
        if from_snap is not None:
            cmd1.extend(['--from-snap', from_snap])
        if src_snap:
            path = utils.convert_str("%s/%s@%s"
                                     % (src_pool, src_name, src_snap))
        else:
            path = utils.convert_str("%s/%s" % (src_pool, src_name))
        cmd1.extend([path, '-'])

        cmd2 = ['rbd', 'import-diff'] + dest_ceph_args
        rbd_path = utils.convert_str("%s/%s" % (dest_pool, dest_name))
        cmd2.extend(['-', rbd_path])

        ret, stderr = self._piped_execute(cmd1, cmd2)
        if ret:
            msg = (_("RBD diff op failed - (ret=%(ret)s stderr=%(stderr)s)") %
                   {'ret': ret, 'stderr': stderr})
            LOG.info(msg)
            raise exception.BackupRBDOperationFailed(msg)

    def _rbd_image_exists(self, name, volume_id, client,
                          try_diff_format=False):
        """Return tuple (exists, name)."""
        rbds = eventlet.tpool.Proxy(self.rbd.RBD()).list(client.ioctx)
        if name not in rbds:
            LOG.debug("Image '%s' not found - trying diff format name", name)
            if try_diff_format:
                name = self._get_backup_base_name(volume_id)
                if name not in rbds:
                    LOG.debug("Diff format image '%s' not found", name)
                    return False, name
            else:
                return False, name

        return True, name

    def _snap_exists(self, base_name, snap_name, client):
        """Return True if snapshot exists in base image."""
        base_rbd = eventlet.tpool.Proxy(self.rbd.Image(client.ioctx,
                                        base_name, read_only=True))
        try:
            snaps = base_rbd.list_snaps()

            if snaps is None:
                return False

            for snap in snaps:
                if snap['name'] == snap_name:
                    return True
        finally:
            base_rbd.close()

        return False

    def _full_rbd_backup(self, container, base_name, length):
        """Create the base_image for a full RBD backup."""
        with eventlet.tpool.Proxy(rbd_driver.RADOSClient(self,
                                  container)) as client:
            self._create_base_image(base_name, length, client)
        # Now we just need to return from_snap=None and image_created=True, if
        # there is some exception in making backup snapshot, will clean up the
        # base image.
        return None, True

    def _incremental_rbd_backup(self, backup, base_name, length,
                                source_rbd_image, volume_id):
        """Select the last snapshot for a RBD incremental backup."""

        container = backup.container
        last_incr = backup.parent_id
        LOG.debug("Trying to perform an incremental backup with container: "
                  "%(container)s, base_name: %(base)s, source RBD image: "
                  "%(source)s, volume ID %(volume)s and last incremental "
                  "backup ID: %(incr)s.",
                  {'container': container,
                   'base': base_name,
                   'source': source_rbd_image,
                   'volume': volume_id,
                   'incr': last_incr,
                   })

        with eventlet.tpool.Proxy(rbd_driver.RADOSClient(self,
                                  container)) as client:
            base_rbd = eventlet.tpool.Proxy(self.rbd.Image(client.ioctx,
                                                           base_name,
                                                           read_only=True))
            try:
                from_snap = self._get_backup_snap_name(base_rbd,
                                                       base_name,
                                                       last_incr)
                if from_snap is None:
                    msg = (_(
                        "Can't find snapshot from parent %(incr)s and "
                        "base name image %(base)s.") %
                        {'incr': last_incr, 'base': base_name})
                    LOG.error(msg)
                    raise exception.BackupRBDOperationFailed(msg)
            finally:
                base_rbd.close()

        return from_snap, False

    def _backup_rbd(self, backup, volume_file, volume_name, length):
        """Create an incremental or full backup from an RBD image."""
        rbd_user = volume_file.rbd_user
        rbd_pool = volume_file.rbd_pool
        rbd_conf = volume_file.rbd_conf
        source_rbd_image = eventlet.tpool.Proxy(volume_file.rbd_image)
        volume_id = backup.volume_id
        base_name = None

        # If backup.parent_id is None performs full RBD backup
        if backup.parent_id is None:
            base_name = self._get_backup_base_name(volume_id, backup=backup)
            from_snap, image_created = self._full_rbd_backup(backup.container,
                                                             base_name,
                                                             length)
        # Otherwise performs incremental rbd backup
        else:
            # Find the base name from the parent backup's service_metadata
            base_name = self._get_backup_base_name(volume_id, backup=backup)
            rbd_img = source_rbd_image
            from_snap, image_created = self._incremental_rbd_backup(backup,
                                                                    base_name,
                                                                    length,
                                                                    rbd_img,
                                                                    volume_id)

        LOG.debug("Using --from-snap '%(snap)s' for incremental backup of "
                  "volume %(volume)s.",
                  {'snap': from_snap, 'volume': volume_id})

        # Snapshot source volume so that we have a new point-in-time
        new_snap = self._get_new_snap_name(backup.id)
        LOG.debug("Creating backup snapshot='%s'", new_snap)
        source_rbd_image.create_snap(new_snap)

        # Attempt differential backup. If this fails, perhaps because librbd
        # or Ceph cluster version does not support it, do a full backup
        # instead.
        #
        # TODO(dosaboy): find a way to determine if the operation is supported
        #                rather than brute force approach.
        try:
            before = time.time()
            self._rbd_diff_transfer(volume_name, rbd_pool, base_name,
                                    backup.container,
                                    src_user=rbd_user,
                                    src_conf=rbd_conf,
                                    dest_user=self._ceph_backup_user,
                                    dest_conf=self._ceph_backup_conf,
                                    src_snap=new_snap,
                                    from_snap=from_snap)

            LOG.debug("Differential backup transfer completed in %.4fs",
                      (time.time() - before))

        except exception.BackupRBDOperationFailed:
            with excutils.save_and_reraise_exception():
                LOG.debug("Differential backup transfer failed")

                # Clean up if image was created as part of this operation
                if image_created:
                    self._try_delete_base_image(backup, base_name=base_name)

                # Delete snapshot
                LOG.debug("Deleting diff backup snapshot='%(snapshot)s' of "
                          "source volume='%(volume)s'.",
                          {'snapshot': new_snap, 'volume': volume_id})
                source_rbd_image.remove_snap(new_snap)

        return {'service_metadata': '{"base": "%s"}' % base_name}

    def _file_is_rbd(self, volume_file):
        """Returns True if the volume_file is actually an RBD image."""
        return hasattr(volume_file, 'rbd_image')

    def _full_backup(self, backup, src_volume, src_name, length):
        """Perform a full backup of src volume.

        First creates a base backup image in our backup location then performs
        an chunked copy of all data from source volume to a new backup rbd
        image.
        """
        volume_id = backup.volume_id
        backup_name = self._get_backup_base_name(volume_id, backup=backup)

        with eventlet.tpool.Proxy(rbd_driver.RADOSClient(self,
                                  backup.container)) as client:
            # First create base backup image
            old_format, features = self._get_rbd_support()
            LOG.debug("Creating backup base image='%(name)s' for volume "
                      "%(volume)s.",
                      {'name': backup_name, 'volume': volume_id})
            eventlet.tpool.Proxy(self.rbd.RBD()).create(
                ioctx=client.ioctx,
                name=backup_name,
                size=length,
                old_format=old_format,
                features=features,
                stripe_unit=self.rbd_stripe_unit,
                stripe_count=self.rbd_stripe_count)

            LOG.debug("Copying data from volume %s.", volume_id)
            dest_rbd = eventlet.tpool.Proxy(self.rbd.Image(client.ioctx,
                                            backup_name))
            try:
                rbd_meta = linuxrbd.RBDImageMetadata(dest_rbd,
                                                     backup.container,
                                                     self._ceph_backup_user,
                                                     self._ceph_backup_conf)
                rbd_fd = linuxrbd.RBDVolumeIOWrapper(rbd_meta)
                self._transfer_data(src_volume, src_name,
                                    eventlet.tpool.Proxy(rbd_fd),
                                    backup_name, length)
            finally:
                dest_rbd.close()

    @staticmethod
    def backup_snapshot_name_pattern():
        """Returns the pattern used to match backup snapshots.

        It is essential that snapshots created for purposes other than backups
        do not have this name format.
        """
        return r"^backup\.([a-z0-9\-]+?)\.snap\.(.+)$"

    @classmethod
    def get_backup_snaps(cls, rbd_image, sort=False):
        """Get all backup snapshots for the given rbd image.

        NOTE: this call is made public since these snapshots must be deleted
              before the base volume can be deleted.
        """
        snaps = rbd_image.list_snaps()

        backup_snaps = []
        for snap in snaps:
            search_key = cls.backup_snapshot_name_pattern()
            result = re.search(search_key, snap['name'])
            if result:
                backup_snaps.append({'name': result.group(0),
                                     'backup_id': result.group(1),
                                     'timestamp': result.group(2)})

        if sort:
            # Sort into ascending order of timestamp
            backup_snaps.sort(key=lambda x: x['timestamp'], reverse=True)

        return backup_snaps

    def _get_new_snap_name(self, backup_id):
        return utils.convert_str("backup.%s.snap.%s"
                                 % (backup_id, time.time()))

    def _get_backup_snap_name(self, rbd_image, name, backup_id):
        """Return the name of the snapshot associated with backup_id.

        The rbd image provided must be the base image used for an incremental
        backup.

        A backup is only allowed ONE associated snapshot. If more are found,
        exception.BackupOperationError is raised.
        """
        snaps = self.get_backup_snaps(rbd_image)

        LOG.debug("Looking for snapshot of backup base '%s'", name)

        if not snaps:
            LOG.debug("Backup base '%s' has no snapshots", name)
            return None

        snaps = [snap['name'] for snap in snaps
                 if snap['backup_id'] == backup_id]

        if not snaps:
            LOG.debug("Backup '%s' has no snapshot", backup_id)
            return None

        if len(snaps) > 1:
            msg = (_("Backup should only have one snapshot but instead has %s")
                   % len(snaps))
            LOG.error(msg)
            raise exception.BackupOperationError(msg)

        LOG.debug("Found snapshot '%s'", snaps[0])
        return snaps[0]

    def _get_volume_size_bytes(self, volume):
        """Return the size in bytes of the given volume.

        Raises exception.InvalidParameterValue if volume size is 0.
        """
        if int(volume['size']) == 0:
            errmsg = _("Need non-zero volume size")
            raise exception.InvalidParameterValue(errmsg)

        return int(volume['size']) * units.Gi

    def _backup_metadata(self, backup):
        """Backup volume metadata.

        NOTE(dosaboy): the metadata we are backing up is obtained from a
                       versioned api so we should not alter it in any way here.
                       We must also be sure that the service that will perform
                       the restore is compatible with version used.
        """
        json_meta = self.get_metadata(backup.volume_id)
        if not json_meta:
            LOG.debug("No metadata to backup for volume %s.", backup.volume_id)
            return

        LOG.debug("Backing up metadata for volume %s.", backup.volume_id)
        try:
            with eventlet.tpool.Proxy(rbd_driver.RADOSClient(self,
                                      backup.container)) as client:
                vol_meta_backup = VolumeMetadataBackup(client, backup.id)
                vol_meta_backup.set(json_meta)
        except exception.VolumeMetadataBackupExists as e:
            msg = (_("Failed to backup volume metadata - %s") % e)
            raise exception.BackupOperationError(msg)

    def backup(self, backup, volume_file, backup_metadata=True):
        """Backup volume and metadata (if available) to Ceph object store.

        If the source volume is an RBD we will attempt to do an
        incremental/differential backup, otherwise a full copy is performed.
        If this fails we will attempt to fall back to full copy.
        """
        volume = self.db.volume_get(self.context, backup.volume_id)
        updates = {}
        if not backup.container:
            backup.container = self._ceph_backup_pool
            backup.save()

        LOG.debug("Starting backup of volume='%s'.", volume.id)

        # Ensure we are at the beginning of the volume
        volume_file.seek(0)
        length = self._get_volume_size_bytes(volume)

        if backup.snapshot_id:
            do_full_backup = True
        elif self._file_is_rbd(volume_file):
            # If volume an RBD, attempt incremental or full backup.
            do_full_backup = False
            LOG.debug("Volume file is RBD: attempting optimized backup")
            try:
                updates = self._backup_rbd(backup, volume_file, volume.name,
                                           length)
            except exception.BackupRBDOperationFailed:
                with excutils.save_and_reraise_exception():
                    self.delete_backup(backup)
        else:
            if backup.parent_id:
                LOG.debug("Volume file is NOT RBD: can't perform "
                          "incremental backup.")
                raise exception.BackupRBDOperationFailed
            LOG.debug("Volume file is NOT RBD: will do full backup.")
            do_full_backup = True

        if do_full_backup:
            try:
                self._full_backup(backup, volume_file, volume.name, length)
            except exception.BackupOperationError:
                with excutils.save_and_reraise_exception():
                    self.delete_backup(backup)

        if backup_metadata:
            try:
                self._backup_metadata(backup)
            except exception.BackupOperationError:
                with excutils.save_and_reraise_exception():
                    # Cleanup.
                    self.delete_backup(backup)

        LOG.debug("Backup '%(backup_id)s' of volume %(volume_id)s finished.",
                  {'backup_id': backup.id, 'volume_id': volume.id})

        return updates

    def _full_restore(self, backup, dest_file, dest_name, length,
                      src_snap=None):
        """Restore volume using full copy i.e. all extents.

        This will result in all extents being copied from source to
        destination.
        """
        with eventlet.tpool.Proxy(rbd_driver.RADOSClient(self,
                                  backup.container)) as client:
            # If a source snapshot is provided we assume the base is diff
            # format.
            if src_snap:
                backup_name = self._get_backup_base_name(backup.volume_id,
                                                         backup=backup)
            else:
                backup_name = self._get_backup_base_name(backup.volume_id)

            # Retrieve backup volume
            src_rbd = eventlet.tpool.Proxy(self.rbd.Image(client.ioctx,
                                                          backup_name,
                                                          snapshot=src_snap,
                                                          read_only=True))
            try:
                rbd_meta = linuxrbd.RBDImageMetadata(src_rbd,
                                                     backup.container,
                                                     self._ceph_backup_user,
                                                     self._ceph_backup_conf)
                rbd_fd = linuxrbd.RBDVolumeIOWrapper(rbd_meta)
                self._transfer_data(eventlet.tpool.Proxy(rbd_fd), backup_name,
                                    dest_file, dest_name, length)
            finally:
                src_rbd.close()

    def _check_restore_vol_size(self, backup, restore_vol, restore_length,
                                src_pool):
        """Ensure that the restore volume is the correct size.

        If the restore volume was bigger than the backup, the diff restore will
        shrink it to the size of the original backup so we need to
        post-process and resize it back to its expected size.
        """
        backup_base = self._get_backup_base_name(backup.volume_id,
                                                 backup=backup)

        with eventlet.tpool.Proxy(rbd_driver.RADOSClient(self,
                                  backup.container)) as client:
            adjust_size = 0
            base_image = eventlet.tpool.Proxy(self.rbd.Image(client.ioctx,
                                              utils.convert_str(backup_base),
                                              read_only=True))
            try:
                if restore_length != base_image.size():
                    adjust_size = restore_length
            finally:
                base_image.close()

        if adjust_size:
            LOG.debug("Adjusting restore vol size")
            restore_vol.rbd_image.resize(adjust_size)

    def _diff_restore_rbd(self, backup, restore_file, restore_name,
                          restore_point, restore_length):
        """Attempt restore rbd volume from backup using diff transfer."""
        rbd_user = restore_file.rbd_user
        rbd_pool = restore_file.rbd_pool
        rbd_conf = restore_file.rbd_conf
        base_name = self._get_backup_base_name(backup.volume_id,
                                               backup=backup)

        LOG.debug("Attempting incremental restore from base='%(base)s' "
                  "snap='%(snap)s'",
                  {'base': base_name, 'snap': restore_point})
        before = time.time()
        try:
            self._rbd_diff_transfer(base_name, backup.container,
                                    restore_name, rbd_pool,
                                    src_user=self._ceph_backup_user,
                                    src_conf=self._ceph_backup_conf,
                                    dest_user=rbd_user, dest_conf=rbd_conf,
                                    src_snap=restore_point)
        except exception.BackupRBDOperationFailed:
            LOG.exception("Differential restore failed, trying full restore")
            raise

        # If the volume we are restoring to is larger than the backup volume,
        # we will need to resize it after the diff import since import-diff
        # appears to shrink the target rbd volume to the size of the original
        # backup volume.
        self._check_restore_vol_size(backup, restore_file, restore_length,
                                     rbd_pool)

        LOG.debug("Restore transfer completed in %.4fs",
                  (time.time() - before))

    def _get_restore_point(self, base_name, backup_id):
        """Get restore point snapshot name for incremental backup.

        If the backup was not incremental (determined by the fact that the
        base has no snapshots/restore points), None is returned. Otherwise, the
        restore point associated with backup_id is returned.
        """
        with eventlet.tpool.Proxy(rbd_driver.RADOSClient(self,
                                  self._ceph_backup_pool)) as client:
            base_rbd = eventlet.tpool.Proxy(self.rbd.Image(client.ioctx,
                                            base_name, read_only=True))
            try:
                restore_point = self._get_backup_snap_name(base_rbd, base_name,
                                                           backup_id)
            finally:
                base_rbd.close()

        return restore_point

    def _rbd_has_extents(self, rbd_volume):
        """Check whether the given rbd volume has extents.

        Return True if has extents, otherwise False.
        """
        extents = []

        def iter_cb(offset, length, exists):
            if exists:
                extents.append(length)

        rbd_volume.diff_iterate(0, rbd_volume.size(), None, iter_cb)

        if extents:
            LOG.debug("RBD has %s extents", sum(extents))
            return True

        return False

    def _diff_restore_allowed(self, base_name, backup, volume, volume_file,
                              rados_client):
        """Determine if differential restore is possible and restore point.

        Determine whether a differential restore is possible/allowed,
        and find out the restore point if backup base is diff-format.

        In order for a differential restore to be performed we need:
            * destination volume must be RBD
            * destination volume must have zero extents
            * backup base image must exist
            * backup must have a restore point
            * target volume is different from source volume of backup

        Returns True if differential restore is allowed, False otherwise.
        Return the restore point if back base is diff-format.
        """
        # NOTE(dosaboy): base_name here must be diff format.
        rbd_exists, base_name = self._rbd_image_exists(base_name,
                                                       backup.volume_id,
                                                       rados_client)

        if not rbd_exists:
            return False, None

        # Get the restore point. If no restore point is found, we assume
        # that the backup was not performed using diff/incremental methods
        # so we enforce full copy.
        restore_point = self._get_restore_point(base_name, backup.id)

        if restore_point:
            if self._file_is_rbd(volume_file):
                LOG.debug("Volume file is RBD.")
                # If the volume we are restoring to is the volume the backup
                # was made from, force a full restore since a diff will not
                # work in this case.
                if volume.id == backup.volume_id:
                    LOG.debug("Destination volume is same as backup source "
                              "volume %s - forcing full copy.", volume.id)
                    return False, restore_point

                # If the destination volume has extents we cannot allow a diff
                # restore.
                if self._rbd_has_extents(volume_file.rbd_image):
                    # We return the restore point so that a full copy is done
                    # from snapshot.
                    LOG.debug("Destination has extents - forcing full copy")
                    return False, restore_point

                return True, restore_point
            else:
                LOG.debug("Volume file is NOT RBD.")
        else:
            LOG.info("No restore point found for backup='%(backup)s' of "
                     "volume %(volume)s although base image is found - "
                     "forcing full copy.",
                     {'backup': backup.id,
                      'volume': backup.volume_id})
        return False, restore_point

    def _restore_volume(self, backup, volume, volume_file):
        """Restore volume from backup using diff transfer if possible.

        Attempts a differential restore and reverts to full copy if diff fails.
        """
        length = int(volume.size) * units.Gi

        if backup.service_metadata:
            base_name = self._get_backup_base_name(backup.volume_id, backup)
        else:
            base_name = self._get_backup_base_name(backup.volume_id)

        with eventlet.tpool.Proxy(rbd_driver.RADOSClient(
                                  self, backup.container)) as client:
            diff_allowed, restore_point = \
                self._diff_restore_allowed(base_name, backup, volume,
                                           volume_file, client)

        do_full_restore = True
        if diff_allowed:
            # Attempt diff
            try:
                LOG.debug("Attempting differential restore.")
                self._diff_restore_rbd(backup, volume_file, volume.name,
                                       restore_point, length)
                do_full_restore = False
            except exception.BackupRBDOperationFailed:
                LOG.debug("Forcing full restore to volume %s.",
                          volume.id)

        if do_full_restore:
            # Otherwise full copy
            LOG.debug("Running full restore.")
            self._full_restore(backup, volume_file, volume.name,
                               length, src_snap=restore_point)

    def _restore_metadata(self, backup, volume_id):
        """Restore volume metadata from backup.

        If this backup has associated metadata, save it to the restore target
        otherwise do nothing.
        """
        try:
            with eventlet.tpool.Proxy(rbd_driver.RADOSClient(self)) as client:
                meta_bak = VolumeMetadataBackup(client, backup.id)
                meta = meta_bak.get()
                if meta is not None:
                    self.put_metadata(volume_id, meta)
                else:
                    LOG.debug("Volume %s has no backed up metadata.",
                              backup.volume_id)
        except exception.BackupMetadataUnsupportedVersion:
            msg = _("Metadata restore failed due to incompatible version")
            LOG.error(msg)
            raise exception.BackupOperationError(msg)

    def restore(self, backup, volume_id, volume_file):
        """Restore volume from backup in Ceph object store.

        If volume metadata is available this will also be restored.
        """
        target_volume = self.db.volume_get(self.context, volume_id)
        LOG.debug('Starting restore from Ceph backup=%(src)s to '
                  'volume=%(dest)s',
                  {'src': backup.id, 'dest': target_volume.name})

        try:
            self._restore_volume(backup, target_volume, volume_file)

            # Be tolerant of IO implementations that do not support fileno()
            try:
                fileno = volume_file.fileno()
            except IOError:
                LOG.debug("Restore target I/O object does not support "
                          "fileno() - skipping call to fsync().")
            else:
                os.fsync(fileno)

            self._restore_metadata(backup, volume_id)

            LOG.debug('Restore to volume %s finished successfully.',
                      volume_id)
        except exception.BackupOperationError as e:
            LOG.error('Restore to volume %(volume)s finished with error - '
                      '%(error)s.', {'error': e, 'volume': volume_id})
            raise

    def delete_backup(self, backup):
        """Delete the given backup from Ceph object store."""
        LOG.debug('Delete started for backup=%s', backup.id)

        delete_failed = False
        has_pool = True
        try:
            self._try_delete_base_image(backup)
        except self.rbd.ImageNotFound:
            LOG.warning(
                "RBD image for backup %(backup)s of volume %(volume)s "
                "not found. Deleting backup metadata.",
                {'backup': backup.id, 'volume': backup.volume_id})
            delete_failed = True
        except self.rados.ObjectNotFound:
            LOG.warning("The pool %(pool)s doesn't exist.",
                        {'pool': backup.container})
            delete_failed = True
            has_pool = False

        if has_pool:
            with eventlet.tpool.Proxy(rbd_driver.RADOSClient(
                                      self, backup.container)) as client:
                VolumeMetadataBackup(client, backup.id).remove_if_exists()

        if delete_failed:
            LOG.info("Delete of backup '%(backup)s' for volume '%(volume)s' "
                     "finished with warning.",
                     {'backup': backup.id, 'volume': backup.volume_id})
        else:
            LOG.debug("Delete of backup '%(backup)s' for volume "
                      "'%(volume)s' finished.",
                      {'backup': backup.id, 'volume': backup.volume_id})



#    Copyright 2015 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import timeutils

from cinder import context
from cinder import exception
from cinder.objects import base as obj_base
from cinder.tests.unit import test


class BaseObjectsTestCase(test.TestCase):
    def setUp(self, *args, **kwargs):
        super(BaseObjectsTestCase, self).setUp(*args, **kwargs)
        self.user_id = 'fake-user'
        self.project_id = 'fake-project'
        self.context = context.RequestContext(self.user_id, self.project_id,
                                              is_admin=False)
        # We only test local right now.
        # TODO(mriedem): Testing remote would be nice...
        self.assertIsNone(obj_base.CinderObject.indirection_api)

    # TODO(mriedem): Replace this with
    # oslo_versionedobjects.fixture.compare_obj when that is in a released
    # version of o.vo.
    @staticmethod
    def _compare(test, db, obj):
        for field, value in db.items():
            try:
                getattr(obj, field)
            except (AttributeError, exception.CinderException,
                    NotImplementedError):
                # NotImplementedError: ignore "Cannot load 'projects' in the
                # base class" error
                continue

            obj_field = getattr(obj, field)
            if field in ('modified_at', 'created_at', 'updated_at',
                         'deleted_at', 'last_heartbeat') and db[field]:
                test.assertEqual(db[field],
                                 timeutils.normalize_time(obj_field))
            elif isinstance(obj_field, obj_base.ObjectListBase):
                test.assertEqual(db[field], obj_field.objects)
            else:
                test.assertEqual(db[field], obj_field)



# Copyright 2010 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import enum
import json
import os
import re

from oslo_config import cfg
from oslo_log import log as logging
from six.moves import urllib
import webob

from cinder.api import api_utils
from cinder.api import microversions as mv
from cinder.common import constants
from cinder import exception
from cinder.i18n import _


api_common_opts = [
    cfg.IntOpt('osapi_max_limit',
               default=1000,
               help='The maximum number of items that a collection '
                    'resource returns in a single response'),
    cfg.StrOpt('resource_query_filters_file',
               default='/etc/cinder/resource_filters.json',
               help="Json file indicating user visible filter "
                    "parameters for list queries."),
]

CONF = cfg.CONF
CONF.import_opt('public_endpoint', 'cinder.api.views.versions')
CONF.register_opts(api_common_opts)

LOG = logging.getLogger(__name__)
_FILTERS_COLLECTION = None

ATTRIBUTE_CONVERTERS = {'name~': 'display_name~',
                        'description~': 'display_description~'}


METADATA_TYPES = enum.Enum('METADATA_TYPES', 'user image')


def get_pagination_params(params, max_limit=None):
    """Return marker, limit, offset tuple from request.

    :param params: `wsgi.Request`'s GET dictionary, possibly containing
                   'marker',  'limit', and 'offset' variables. 'marker' is the
                   id of the last element the client has seen, 'limit' is the
                   maximum number of items to return and 'offset' is the number
                   of items to skip from the marker or from the first element.
                   If 'limit' is not specified, or > max_limit, we default to
                   max_limit. Negative values for either offset or limit will
                   cause exc.HTTPBadRequest() exceptions to be raised. If no
                   offset is present we'll default to 0 and if no marker is
                   present we'll default to None.
    :max_limit: Max value 'limit' return value can take
    :returns: Tuple (marker, limit, offset)
    """
    max_limit = max_limit or CONF.osapi_max_limit
    limit = _get_limit_param(params, max_limit)
    marker = _get_marker_param(params)
    offset = _get_offset_param(params)
    return marker, limit, offset


def _get_limit_param(params, max_limit=None):
    """Extract integer limit from request's dictionary or fail.

   Defaults to max_limit if not present and returns max_limit if present
   'limit' is greater than max_limit.
    """
    max_limit = max_limit or CONF.osapi_max_limit
    try:
        limit = int(params.pop('limit', max_limit))
    except ValueError:
        msg = _('limit param must be an integer')
        raise webob.exc.HTTPBadRequest(explanation=msg)
    if limit < 0:
        msg = _('limit param must be positive')
        raise webob.exc.HTTPBadRequest(explanation=msg)
    limit = min(limit, max_limit)
    return limit


def _get_marker_param(params):
    """Extract marker id from request's dictionary (defaults to None)."""
    return params.pop('marker', None)


def _get_offset_param(params):
    """Extract offset id from request's dictionary (defaults to 0) or fail."""
    offset = params.pop('offset', 0)
    return api_utils.validate_integer(offset,
                                      'offset',
                                      0,
                                      constants.DB_MAX_INT)


def limited(items, request, max_limit=None):
    """Return a slice of items according to requested offset and limit.

    :param items: A sliceable entity
    :param request: ``wsgi.Request`` possibly containing 'offset' and 'limit'
                    GET variables. 'offset' is where to start in the list,
                    and 'limit' is the maximum number of items to return. If
                    'limit' is not specified, 0, or > max_limit, we default
                    to max_limit. Negative values for either offset or limit
                    will cause exc.HTTPBadRequest() exceptions to be raised.
    :kwarg max_limit: The maximum number of items to return from 'items'
    """
    max_limit = max_limit or CONF.osapi_max_limit
    marker, limit, offset = get_pagination_params(request.GET.copy(),
                                                  max_limit)
    range_end = offset + (limit or max_limit)
    return items[offset:range_end]


def get_sort_params(params, default_key='created_at', default_dir='desc'):
    """Retrieves sort keys/directions parameters.

    Processes the parameters to create a list of sort keys and sort directions
    that correspond to either the 'sort' parameter or the 'sort_key' and
    'sort_dir' parameter values. The value of the 'sort' parameter is a comma-
    separated list of sort keys, each key is optionally appended with
    ':<sort_direction>'.

    Note that the 'sort_key' and 'sort_dir' parameters are deprecated in kilo
    and an exception is raised if they are supplied with the 'sort' parameter.

    The sort parameters are removed from the request parameters by this
    function.

    :param params: webob.multidict of request parameters (from
                   cinder.api.openstack.wsgi.Request.params)
    :param default_key: default sort key value, added to the list if no
                        sort keys are supplied
    :param default_dir: default sort dir value, added to the list if the
                        corresponding key does not have a direction
                        specified
    :returns: list of sort keys, list of sort dirs
    :raise webob.exc.HTTPBadRequest: If both 'sort' and either 'sort_key' or
                                     'sort_dir' are supplied parameters
    """
    if 'sort' in params and ('sort_key' in params or 'sort_dir' in params):
        msg = _("The 'sort_key' and 'sort_dir' parameters are deprecated and "
                "cannot be used with the 'sort' parameter.")
        raise webob.exc.HTTPBadRequest(explanation=msg)
    sort_keys = []
    sort_dirs = []
    if 'sort' in params:
        for sort in params.pop('sort').strip().split(','):
            sort_key, _sep, sort_dir = sort.partition(':')
            if not sort_dir:
                sort_dir = default_dir
            sort_keys.append(sort_key.strip())
            sort_dirs.append(sort_dir.strip())
    else:
        sort_key = params.pop('sort_key', default_key)
        sort_dir = params.pop('sort_dir', default_dir)
        sort_keys.append(sort_key.strip())
        sort_dirs.append(sort_dir.strip())
    return sort_keys, sort_dirs


def get_request_url(request):
    url = request.application_url
    headers = request.headers
    forwarded = headers.get('X-Forwarded-Host')
    if forwarded:
        url_parts = list(urllib.parse.urlsplit(url))
        url_parts[1] = re.split(r',\s?', forwarded)[-1]
        url = urllib.parse.urlunsplit(url_parts).rstrip('/')
    return url


def remove_version_from_href(href):
    """Removes the first API version from the href.

    Given: 'http://cinder.example.com/v1.1/123'
    Returns: 'http://cinder.example.com/123'

    Given: 'http://cinder.example.com/v1.1'
    Returns: 'http://cinder.example.com'

    Given: 'http://cinder.example.com/volume/drivers/v1.1/flashsystem'
    Returns: 'http://cinder.example.com/volume/drivers/flashsystem'

    """
    parsed_url = urllib.parse.urlsplit(href)
    url_parts = parsed_url.path.split('/')

    # NOTE: this should match vX.X or vX
    expression = re.compile(r'^v([0-9]+|[0-9]+\.[0-9]+)(/.*|$)')
    for x in range(len(url_parts)):
        if expression.match(url_parts[x]):
            del url_parts[x]
            break

    new_path = '/'.join(url_parts)

    if new_path == parsed_url.path:
        msg = 'href %s does not contain version' % href
        LOG.debug(msg)
        raise ValueError(msg)

    parsed_url = list(parsed_url)
    parsed_url[2] = new_path
    return urllib.parse.urlunsplit(parsed_url)


class ViewBuilder(object):
    """Model API responses as dictionaries."""

    _collection_name = None

    def _get_links(self, request, identifier):
        return [{"rel": "self",
                 "href": self._get_href_link(request, identifier), },
                {"rel": "bookmark",
                 "href": self._get_bookmark_link(request, identifier), }]

    def _get_next_link(self, request, identifier, collection_name):
        """Return href string with proper limit and marker params."""
        params = request.params.copy()
        params["marker"] = identifier
        prefix = self._update_link_prefix(get_request_url(request),
                                          CONF.public_endpoint)
        url = os.path.join(prefix,
                           request.environ["cinder.context"].project_id,
                           collection_name)
        return "%s?%s" % (url, urllib.parse.urlencode(params))

    def _get_href_link(self, request, identifier):
        """Return an href string pointing to this object."""
        prefix = self._update_link_prefix(get_request_url(request),
                                          CONF.public_endpoint)
        return os.path.join(prefix,
                            request.environ["cinder.context"].project_id,
                            self._collection_name,
                            str(identifier))

    def _get_bookmark_link(self, request, identifier):
        """Create a URL that refers to a specific resource."""
        base_url = remove_version_from_href(get_request_url(request))
        base_url = self._update_link_prefix(base_url,
                                            CONF.public_endpoint)
        return os.path.join(base_url,
                            request.environ["cinder.context"].project_id,
                            self._collection_name,
                            str(identifier))

    def _get_collection_links(self, request, items, collection_name,
                              item_count=None, id_key="uuid"):
        """Retrieve 'next' link, if applicable.

        The next link is included if we are returning as many items as we can,
        given the restrictions of limit optional request parameter and
        osapi_max_limit configuration parameter as long as we are returning
        some elements.

        So we return next link if:

        1) 'limit' param is specified and equal to the number of items.
        2) 'limit' param is NOT specified and the number of items is
           equal to CONF.osapi_max_limit.

        :param request: API request
        :param items: List of collection items
        :param collection_name: Name of collection, used to generate the
                                next link for a pagination query
        :param item_count: Length of the list of the original collection
                           items
        :param id_key: Attribute key used to retrieve the unique ID, used
                       to generate the next link marker for a pagination query
        :returns: links
        """
        item_count = item_count or len(items)
        limit = _get_limit_param(request.GET.copy())
        if len(items) and limit <= item_count:
            return self._generate_next_link(items, id_key, request,
                                            collection_name)

        return []

    def _generate_next_link(self, items, id_key, request,
                            collection_name):
        links = []
        last_item = items[-1]
        if id_key in last_item:
            last_item_id = last_item[id_key]
        else:
            last_item_id = last_item["id"]
        links.append({
            "rel": "next",
            "href": self._get_next_link(request, last_item_id,
                                        collection_name),
        })
        return links

    def _update_link_prefix(self, orig_url, prefix):
        if not prefix:
            return orig_url
        url_parts = list(urllib.parse.urlsplit(orig_url))
        prefix_parts = list(urllib.parse.urlsplit(prefix))
        url_parts[0:2] = prefix_parts[0:2]
        url_parts[2] = prefix_parts[2] + url_parts[2]

        return urllib.parse.urlunsplit(url_parts).rstrip('/')


def get_cluster_host(req, params, cluster_version=None):
    """Get cluster and host from the parameters.

    This method checks the presence of cluster and host parameters and returns
    them depending on the cluster_version.

    If cluster_version is False we will never return the cluster_name and we
    will require the presence of the host parameter.

    If cluster_version is None we will always check for the presence of the
    cluster parameter, and if cluster_version is a string with a version we
    will only check for the presence of the parameter if the version of the
    request is not less than  it.  In both cases we will require one and only
    one parameter, host or cluster.
    """
    if (cluster_version is not False and
            req.api_version_request.matches(cluster_version)):
        cluster_name = params.get('cluster')
        msg = _('One and only one of cluster and host must be set.')
    else:
        cluster_name = None
        msg = _('Host field is missing.')

    host = params.get('host')
    if bool(cluster_name) == bool(host):
        raise exception.InvalidInput(reason=msg)
    return cluster_name, host


def _initialize_filters():
    global _FILTERS_COLLECTION
    if _FILTERS_COLLECTION:
        return

    if not os.path.exists(CONF.resource_query_filters_file):
        LOG.error(
            "resource query filters file does not exist: %s",
            CONF.resource_query_filters_file)
        return

    with open(CONF.resource_query_filters_file, 'r') as filters_file:
        _FILTERS_COLLECTION = json.load(filters_file)


def get_enabled_resource_filters(resource=None):
    """Get list of configured/allowed filters for the specified resource.

    This method checks resource_query_filters_file and returns dictionary
    which contains the specified resource and its allowed filters:

    .. code-block:: json

            {
                "resource": ["filter1", "filter2", "filter3"]
            }

    if resource is not specified, all of the configuration will be returned,
    and if the resource is not found, empty dict will be returned.
    """
    try:
        _initialize_filters()
        if not resource:
            return _FILTERS_COLLECTION
        else:
            return {resource: _FILTERS_COLLECTION[resource]}
    except Exception:
        LOG.debug("Failed to collect resource %s's filters.", resource)
        return {}


def get_time_comparsion_operators():
    """Get list of time comparsion operators.

    This method returns list which contains the allowed comparsion operators.
    """
    return ["gt", "gte", "eq", "neq", "lt", "lte"]


def convert_filter_attributes(filters, resource):
    for key in filters.copy().keys():
        if resource in ['volume', 'backup',
                        'snapshot'] and key in ATTRIBUTE_CONVERTERS.keys():
            filters[ATTRIBUTE_CONVERTERS[key]] = filters[key]
            filters.pop(key)


def reject_invalid_filters(context, filters, resource,
                           enable_like_filter=False):
    invalid_filters = []
    for key in filters.copy().keys():
        try:
            # Only ASCII characters can be valid filter keys,
            # in PY2/3, the key can be either unicode or string.
            if isinstance(key, str):
                key.encode('ascii')
            else:
                key.decode('ascii')
        except (UnicodeEncodeError, UnicodeDecodeError):
            raise webob.exc.HTTPBadRequest(
                explanation=_('Filter keys can only contain '
                              'ASCII characters.'))

    if context.is_admin and resource not in ['pool']:
        # Allow all options except resource is pool
        # pool API is only available for admin
        return
    # Check the configured filters against those passed in resource
    configured_filters = get_enabled_resource_filters(resource)
    if configured_filters:
        configured_filters = configured_filters[resource]
    else:
        configured_filters = []
    for key in filters.copy().keys():
        if not enable_like_filter:
            if key not in configured_filters:
                invalid_filters.append(key)
        else:
            # If 'key~' is configured, both 'key' and 'key~' are valid.
            if not (key in configured_filters or
                    "%s~" % key in configured_filters):
                invalid_filters.append(key)
    if invalid_filters:
        if 'all_tenants' in invalid_filters:
            invalid_filters.remove('all_tenants')
        if len(invalid_filters) == 0:
            return
        raise webob.exc.HTTPBadRequest(
            explanation=_('Invalid filters %s are found in query '
                          'options.') % ','.join(invalid_filters))


def process_general_filtering(resource):
    def wrapper(process_non_general_filtering):
        def _decorator(*args, **kwargs):
            req_version = kwargs.get('req_version')
            filters = kwargs.get('filters')
            context = kwargs.get('context')
            if req_version.matches(mv.RESOURCE_FILTER):
                support_like = False
                if req_version.matches(mv.LIKE_FILTER):
                    support_like = True
                reject_invalid_filters(context, filters,
                                       resource, support_like)
                convert_filter_attributes(filters, resource)

            else:
                process_non_general_filtering(*args, **kwargs)
        return _decorator
    return wrapper



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The volume type access extension."""

import six
from six.moves import http_client
import webob

from cinder.api import extensions
from cinder.api.openstack import wsgi
from cinder.api.schemas import volume_type_access
from cinder.api import validation
from cinder import exception
from cinder.i18n import _
from cinder.policies import volume_access as policy
from cinder.volume import volume_types


def _marshall_volume_type_access(vol_type):
    rval = []
    for project_id in vol_type['projects']:
        rval.append({'volume_type_id': vol_type['id'],
                     'project_id': project_id})

    return {'volume_type_access': rval}


class VolumeTypeAccessController(object):
    """The volume type access API controller for the OpenStack API."""

    def index(self, req, type_id):
        context = req.environ['cinder.context']
        context.authorize(policy.TYPE_ACCESS_POLICY)

        # Not found exception will be handled at the wsgi level
        vol_type = volume_types.get_volume_type(
            context, type_id, expected_fields=['projects'])

        if vol_type['is_public']:
            expl = _("Access list not available for public volume types.")
            raise exception.VolumeTypeAccessNotFound(message=expl)

        return _marshall_volume_type_access(vol_type)


class VolumeTypeActionController(wsgi.Controller):
    """The volume type access API controller for the OpenStack API."""

    def _extend_vol_type(self, vol_type_rval, vol_type_ref):
        if vol_type_ref:
            key = "%s:is_public" % (Volume_type_access.alias)
            vol_type_rval[key] = vol_type_ref.get('is_public', True)

    @wsgi.extends
    def show(self, req, resp_obj, id):
        context = req.environ['cinder.context']
        if context.authorize(policy.TYPE_ACCESS_POLICY, fatal=False):
            vol_type = req.cached_resource_by_id(id, name='types')
            self._extend_vol_type(resp_obj.obj['volume_type'], vol_type)

    @wsgi.extends
    def index(self, req, resp_obj):
        context = req.environ['cinder.context']
        if context.authorize(policy.TYPE_ACCESS_POLICY, fatal=False):
            for vol_type_rval in list(resp_obj.obj['volume_types']):
                type_id = vol_type_rval['id']
                vol_type = req.cached_resource_by_id(type_id, name='types')
                self._extend_vol_type(vol_type_rval, vol_type)

    @wsgi.extends
    def detail(self, req, resp_obj):
        context = req.environ['cinder.context']
        if context.authorize(policy.TYPE_ACCESS_POLICY, fatal=False):
            for vol_type_rval in list(resp_obj.obj['volume_types']):
                type_id = vol_type_rval['id']
                vol_type = req.cached_resource_by_id(type_id, name='types')
                self._extend_vol_type(vol_type_rval, vol_type)

    @wsgi.extends(action='create')
    def create(self, req, body, resp_obj):
        context = req.environ['cinder.context']
        if context.authorize(policy.TYPE_ACCESS_POLICY, fatal=False):
            type_id = resp_obj.obj['volume_type']['id']
            vol_type = req.cached_resource_by_id(type_id, name='types')
            self._extend_vol_type(resp_obj.obj['volume_type'], vol_type)

    @wsgi.action('addProjectAccess')
    @validation.schema(volume_type_access.add_project_access)
    def _addProjectAccess(self, req, id, body):
        context = req.environ['cinder.context']
        context.authorize(policy.ADD_PROJECT_POLICY)
        project = body['addProjectAccess']['project']

        try:
            volume_types.add_volume_type_access(context, id, project)
        # Not found exception will be handled at the wsgi level
        except exception.VolumeTypeAccessExists as err:
            raise webob.exc.HTTPConflict(explanation=six.text_type(err))
        return webob.Response(status_int=http_client.ACCEPTED)

    @wsgi.action('removeProjectAccess')
    @validation.schema(volume_type_access.remove_project_access)
    def _removeProjectAccess(self, req, id, body):
        context = req.environ['cinder.context']
        context.authorize(policy.REMOVE_PROJECT_POLICY)
        project = body['removeProjectAccess']['project']

        # Not found exception will be handled at the wsgi level
        volume_types.remove_volume_type_access(context, id, project)
        return webob.Response(status_int=http_client.ACCEPTED)


class Volume_type_access(extensions.ExtensionDescriptor):
    """Volume type access support."""

    name = "VolumeTypeAccess"
    alias = "os-volume-type-access"
    updated = "2014-06-26T00:00:00Z"

    def get_resources(self):
        resources = []
        res = extensions.ResourceExtension(
            Volume_type_access.alias,
            VolumeTypeAccessController(),
            parent=dict(member_name='type', collection_name='types'))
        resources.append(res)
        return resources

    def get_controller_extensions(self):
        controller = VolumeTypeActionController()
        extension = extensions.ControllerExtension(self, 'types', controller)
        return [extension]



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests For PickledScheduler.
"""

import datetime

from oslo_serialization import jsonutils
import six

from cinder.scheduler import scheduler_options
from cinder.tests.unit import test


class FakeSchedulerOptions(scheduler_options.SchedulerOptions):
    def __init__(self, last_checked, now, file_old, file_now, data, filedata):
        super(FakeSchedulerOptions, self).__init__()
        # Change internals ...
        self.last_modified = file_old
        self.last_checked = last_checked
        self.data = data

        # For overrides ...
        self._time_now = now
        self._file_now = file_now
        self._file_data = filedata

        self.file_was_loaded = False

    def _get_file_timestamp(self, filename):
        return self._file_now

    def _get_file_handle(self, filename):
        self.file_was_loaded = True
        return six.StringIO(self._file_data)

    def _get_time_now(self):
        return self._time_now


class SchedulerOptionsTestCase(test.TestCase):
    def test_get_configuration_first_time_no_flag(self):
        last_checked = None
        now = datetime.datetime(2012, 1, 1, 1, 1, 1)
        file_old = None
        file_now = datetime.datetime(2012, 1, 1, 1, 1, 1)

        data = dict(a=1, b=2, c=3)
        jdata = jsonutils.dumps(data)

        fake = FakeSchedulerOptions(last_checked, now, file_old, file_now,
                                    {}, jdata)
        self.assertEqual({}, fake.get_configuration())
        self.assertFalse(fake.file_was_loaded)

    def test_get_configuration_first_time_empty_file(self):
        last_checked = None
        now = datetime.datetime(2012, 1, 1, 1, 1, 1)
        file_old = None
        file_now = datetime.datetime(2012, 1, 1, 1, 1, 1)

        jdata = ""

        fake = FakeSchedulerOptions(last_checked, now, file_old, file_now,
                                    {}, jdata)
        self.assertEqual({}, fake.get_configuration('foo.json'))
        self.assertTrue(fake.file_was_loaded)

    def test_get_configuration_first_time_happy_day(self):
        last_checked = None
        now = datetime.datetime(2012, 1, 1, 1, 1, 1)
        file_old = None
        file_now = datetime.datetime(2012, 1, 1, 1, 1, 1)

        data = dict(a=1, b=2, c=3)
        jdata = jsonutils.dumps(data)

        fake = FakeSchedulerOptions(last_checked, now, file_old, file_now,
                                    {}, jdata)
        self.assertEqual(data, fake.get_configuration('foo.json'))
        self.assertTrue(fake.file_was_loaded)

    def test_get_configuration_second_time_no_change(self):
        last_checked = datetime.datetime(2011, 1, 1, 1, 1, 1)
        now = datetime.datetime(2012, 1, 1, 1, 1, 1)
        file_old = datetime.datetime(2012, 1, 1, 1, 1, 1)
        file_now = datetime.datetime(2012, 1, 1, 1, 1, 1)

        data = dict(a=1, b=2, c=3)
        jdata = jsonutils.dumps(data)

        fake = FakeSchedulerOptions(last_checked, now, file_old, file_now,
                                    data, jdata)
        self.assertEqual(data, fake.get_configuration('foo.json'))
        self.assertFalse(fake.file_was_loaded)

    def test_get_configuration_second_time_too_fast(self):
        last_checked = datetime.datetime(2011, 1, 1, 1, 1, 1)
        now = datetime.datetime(2011, 1, 1, 1, 1, 2)
        file_old = datetime.datetime(2012, 1, 1, 1, 1, 1)
        file_now = datetime.datetime(2013, 1, 1, 1, 1, 1)

        old_data = dict(a=1, b=2, c=3)
        data = dict(a=11, b=12, c=13)
        jdata = jsonutils.dumps(data)

        fake = FakeSchedulerOptions(last_checked, now, file_old, file_now,
                                    old_data, jdata)
        self.assertEqual(old_data, fake.get_configuration('foo.json'))
        self.assertFalse(fake.file_was_loaded)

    def test_get_configuration_second_time_change(self):
        last_checked = datetime.datetime(2011, 1, 1, 1, 1, 1)
        now = datetime.datetime(2012, 1, 1, 1, 1, 1)
        file_old = datetime.datetime(2012, 1, 1, 1, 1, 1)
        file_now = datetime.datetime(2013, 1, 1, 1, 1, 1)

        old_data = dict(a=1, b=2, c=3)
        data = dict(a=11, b=12, c=13)
        jdata = jsonutils.dumps(data)

        fake = FakeSchedulerOptions(last_checked, now, file_old, file_now,
                                    old_data, jdata)
        self.assertEqual(data, fake.get_configuration('foo.json'))
        self.assertTrue(fake.file_was_loaded)



#  Copyright 2012 Pedro Navarro Perez
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Stubouts, mocks and fixtures for windows volume test suite
"""


def get_fake_volume_info():
    return {'name': 'volume_name',
            'size': 1,
            'provider_location': 'iqn.2010-10.org.openstack:' + 'volume_name',
            'id': 1,
            'provider_auth': None}


def get_fake_volume_info_cloned():
    return {'name': 'volume_name_cloned',
            'size': 1,
            'provider_location': 'iqn.2010-10.org.openstack:' +
                                 'volume_name_cloned',
            'id': 1,
            'provider_auth': None}


def get_fake_image_meta():
    return {'id': '10958016-e196-42e3-9e7f-5d8927ae3099'
            }


def get_fake_snapshot_info():
    return {'name': 'snapshot_name',
            'volume_name': 'volume_name', }


def get_fake_connector_info():
    return {'initiator': 'iqn.2010-10.org.openstack:' + 'volume_name', }



# Copyright (c) 2018-2019 LINBIT HA Solutions GmbH
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_config import cfg
from oslo_utils import timeutils

from cinder import exception as cinder_exception
from cinder.tests.unit import test
from cinder.volume import configuration as conf
from cinder.volume.drivers import linstordrv as drv

CONF = cfg.CONF

CINDER_UNKNOWN = 'unknown'
DISKLESS = 'DISKLESS'
LVM = 'LVM'
LVM_THIN = 'LVM_THIN'
ZFS = 'ZFS'
ZFS_THIN = 'ZFS_THIN'
DRIVER = 'cinder.volume.drivers.linstordrv.'

RESOURCE = {
    'name': 'CV_0348a7d3-3bb9-452d-9f40-2cf5ebfe9131',
    'volume': {
        'device_path': '/dev/drbd1000'
    }
}

RESOURCE_LIST = [{
    'layer_object': {
        'children': [{
            'storage': {
                'storage_volumes': [{
                    'allocated_size_kib': 1048576,
                    'device_path':
                    '/dev/vol/CV_0348a7d3-3bb9-452d-9f40-2cf5ebfe9131_00000',
                    'disk_state': '[]',
                    'usable_size_kib': 1048576,
                    'volume_number': 0}]},
            'type': 'STORAGE'}],
        'drbd': {
            'al_size': 32,
            'al_stripes': 1,
            'drbd_resource_definition': {
                'al_stripe_size_kib': 32,
                'al_stripes': 1,
                'down': False,
                'peer_slots': 7,
                'port': 7005,
                'secret': 'poQZ0Ad/Bq8DT9fA7ydB',
                'transport_type': 'IP'},
            'drbd_volumes': [{
                'allocated_size_kib': 1044740,
                'backing_device':
                    '/dev/vol/CV_0348a7d3-3bb9-452d-9f40-2cf5ebfe9131_00000',
                'device_path': '/dev/drbd1005',
                'drbd_volume_definition': {
                    'minor_number': 1005,
                    'volume_number': 0},
                'usable_size_kib': 1044480}],
            'node_id': 0,
            'peer_slots': 7},
        'type': 'DRBD'},
    'name': 'CV_0348a7d3-3bb9-452d-9f40-2cf5ebfe9131',
    'node_name': 'node-2',
    'state': {'in_use': False},
    'uuid': 'a4ab4670-c5fc-4590-a3a2-39c4685c8c32',
    'volumes': [{
        'allocated_size_kib': 45403,
        'device_path': '/dev/drbd1005',
        'layer_data_list': [{
            'data': {
                'allocated_size_kib': 1044740,
                'backing_device':
                '/dev/vol/CV_0348a7d3-3bb9-452d-9f40-2cf5ebfe9131_00000',
                'device_path': '/dev/drbd1005',
                'drbd_volume_definition': {
                    'minor_number': 1005,
                    'volume_number': 0},
                'usable_size_kib': 1044480},
            'type': 'DRBD'}, {
            'data': {
                'allocated_size_kib': 1048576,
                'device_path':
                '/dev/vol/CV_0348a7d3-3bb9-452d-9f40-2cf5ebfe9131_00000',
                'disk_state': '[]',
                'usable_size_kib': 1048576,
                'volume_number': 0},
            'type': 'STORAGE'}
        ],
        'props': {
            'RestoreFromResource': 'CV_123a2fdc-365f-472e-bb8e-484788712abc',
            'RestoreFromSnapshot': 'SN_68edb708-48de-4da1-9953-b9de9da9f1b8'
        },
        'provider_kind': 'LVM_THIN',
        'state': {'disk_state': 'UpToDate'},
        'storage_pool_name': 'DfltStorPool',
        'uuid': 'e270ba0c-b284-4f21-85cc-602f132a2251',
        'volume_number': 0}]}, {
    'flags': ['DISKLESS'],
    'layer_object': {
        'children': [{
            'storage': {
                'storage_volumes': [{
                    'allocated_size_kib': 0,
                    'usable_size_kib': 1044480,
                    'volume_number': 0}]},
            'type': 'STORAGE'}],
        'drbd': {
            'al_size': 32,
            'al_stripes': 1,
            'drbd_resource_definition': {
                'al_stripe_size_kib': 32,
                'al_stripes': 1,
                'down': False,
                'peer_slots': 7,
                'port': 7005,
                'secret': 'poQZ0Ad/Bq8DT9fA7ydB',
                'transport_type': 'IP'},
            'drbd_volumes': [{
                'allocated_size_kib': 1044740,
                'device_path': '/dev/drbd1005',
                'drbd_volume_definition': {
                    'minor_number': 1005,
                    'volume_number': 0},
                'usable_size_kib': 1044480}],
            'flags': ['DISKLESS'],
            'node_id': 1,
            'peer_slots': 7},
        'type': 'DRBD'},
    'name': 'CV_0348a7d3-3bb9-452d-9f40-2cf5ebfe9131',
    'node_name': 'node-1',
    'state': {'in_use': False},
    'uuid': '11e853df-6f66-4cd9-9fbc-f3f7cc98d5cf',
    'volumes': [{
        'allocated_size_kib': 45403,
        'device_path': '/dev/drbd1005',
        'layer_data_list': [
            {
                'data': {
                    'allocated_size_kib': 1044740,
                    'device_path': '/dev/drbd1005',
                    'drbd_volume_definition': {
                        'minor_number': 1005,
                        'volume_number': 0},
                    'usable_size_kib': 1044480},
                'type': 'DRBD'
            },
            {
                'data': {
                    'allocated_size_kib': 0,
                    'usable_size_kib': 1044480,
                    'volume_number': 0
                },
                'type': 'STORAGE'
            }
        ],
        'provider_kind': 'DISKLESS',
        'state': {'disk_state': 'Diskless'},
        'storage_pool_name': 'DfltStorPool',
        'uuid': '27b4aeec-2b42-41c9-b186-86afc8778046',
        'volume_number': 0
    }]}]

RESOURCE_LIST_RESP = ['node-1', 'node-2']

SNAPSHOT_LIST_RESP = ['node-1']

DISKLESS_LIST_RESP = ['node-1']

RESOURCE_DFN_LIST = [{
    'layer_data': [
        {
            'data': {
                'al_stripe_size_kib': 32,
                'al_stripes': 1,
                'down': False,
                'peer_slots': 7,
                'port': 7005,
                'secret': 'poQZ0Ad/Bq8DT9fA7ydB',
                'transport_type': 'IP'
            },
            'type': 'DRBD'
        },
        {
            'type': 'STORAGE'
        }
    ],
    'name': 'CV_0348a7d3-3bb9-452d-9f40-2cf5ebfe9131',
    'props': {'DrbdPrimarySetOn': 'node-1'},
    'uuid': '9a684294-6db4-40c8-bfeb-e5351200b9db'
}]

RESOURCE_DFN_LIST_RESP = [{
    'rd_name': u'CV_0348a7d3-3bb9-452d-9f40-2cf5ebfe9131',
    'rd_uuid': u'9a684294-6db4-40c8-bfeb-e5351200b9db',
}]

NODES_LIST = [
    {
        'connection_status': 'ONLINE',
        'name': 'node-1',
        'net_interfaces': [{
            'address': '192.168.8.63',
            'name': 'default',
            'satellite_encryption_type': 'PLAIN',
            'satellite_port': 3366,
            'uuid': '9c5b727f-0c62-4040-9a33-96a4fd4aaac3'}],
        'props': {'CurStltConnName': 'default'},
        'type': 'COMBINED',
        'uuid': '69b88ffb-50d9-4576-9843-d7bf4724d043'
    },
    {
        'connection_status': 'ONLINE',
        'name': 'node-2',
        'net_interfaces': [{
            'address': '192.168.8.102',
            'name': 'default',
            'satellite_encryption_type': 'PLAIN',
            'satellite_port': 3366,
            'uuid': '3f911fc9-4f9b-4155-b9da-047d5242484c'}],
        'props': {'CurStltConnName': 'default'},
        'type': 'SATELLITE',
        'uuid': '26bde754-0f05-499c-a63c-9f4e5f30556e'
    }
]

NODES_RESP = [
    {'node_address': '192.168.8.63', 'node_name': 'node-1'},
    {'node_address': '192.168.8.102', 'node_name': 'node-2'}
]

STORAGE_POOL_DEF = [{'storage_pool_name': 'DfltStorPool'}]

STORAGE_POOL_DEF_RESP = ['DfltStorPool']

STORAGE_POOL_LIST = [
    {
        'free_capacity': 104815656,
        'free_space_mgr_name': 'node-2:DfltStorPool',
        'node_name': 'node-2',
        'props': {
            'StorDriver/LvmVg': 'vol',
            'StorDriver/ThinPool': 'thin_pool'
        },
        'provider_kind': 'LVM_THIN',
        'static_traits': {
            'Provisioning': 'Thin',
            'SupportsSnapshots': 'true'
        },
        'storage_pool_name': 'DfltStorPool',
        'total_capacity': 104857600,
        'uuid': '004faf29-be1a-4d74-9470-038bcee2c611'
    },
    {
        'free_capacity': 9223372036854775807,
        'free_space_mgr_name': 'node-1:DfltStorPool',
        'node_name': 'node-1',
        'provider_kind': 'DISKLESS',
        'static_traits': {'SupportsSnapshots': 'false'},
        'storage_pool_name': 'DfltStorPool',
        'total_capacity': 9223372036854775807,
        'uuid': '897da09e-1316-45c0-a308-c07008af42df'
    }
]

STORAGE_POOL_LIST_RESP = [
    {
        'driver_name': 'LVM_THIN',
        'node_name': 'node-2',
        'sp_uuid': '004faf29-be1a-4d74-9470-038bcee2c611',
        'sp_cap': 100.0,
        'sp_free': 100,
        'sp_name': u'DfltStorPool'
    },
    {
        'driver_name': 'DISKLESS',
        'node_name': 'node-1',
        'sp_uuid': '897da09e-1316-45c0-a308-c07008af42df',
        'sp_allocated': 0.0,
        'sp_cap': -1.0,
        'sp_free': -1.0,
        'sp_name': 'DfltStorPool'
    }
]

VOLUME_STATS_RESP = {
    'driver_version': '0.0.7',
    'pools': [{
        'QoS_support': False,
        'backend_state': 'up',
        'filter_function': None,
        'free_capacity_gb': 100,
        'goodness_function': None,
        'location_info': 'linstor://localhost',
        'max_over_subscription_ratio': 0,
        'multiattach': False,
        'pool_name': 'lin-test-driver',
        'provisioned_capacity_gb': 0.0,
        'reserved_percentage': 0,
        'thick_provisioning_support': False,
        'thin_provisioning_support': True,
        'total_capacity_gb': 100.0,
        'total_volumes': 1,
    }],
    'vendor_name': 'LINBIT',
    'volume_backend_name': 'lin-test-driver'
}

CINDER_VOLUME = {
    'id': '0348a7d3-3bb9-452d-9f40-2cf5ebfe9131',
    'name': 'test-lin-vol',
    'size': 1,
    'volume_type_id': 'linstor',
    'created_at': timeutils.utcnow()
}

SNAPSHOT = {
    'id': '0348a7d3-3bb9-452d-9f40-2cf5ebfe9131',
    'volume_id': '0348a7d3-3bb9-452d-9f40-2cf5ebfe9131',
    'volume_size': 1
}

VOLUME_NAMES = {
    'linstor': 'CV_0348a7d3-3bb9-452d-9f40-2cf5ebfe9131',
    'cinder': '0348a7d3-3bb9-452d-9f40-2cf5ebfe9131',
    'snap': 'SN_0348a7d3-3bb9-452d-9f40-2cf5ebfe9131',
}


class LinstorAPIFakeDriver(object):

    def fake_api_ping(self):
        return 1234

    def fake_api_resource_list(self):
        return RESOURCE_LIST

    def fake_api_node_list(self):
        return NODES_LIST

    def fake_api_storage_pool_dfn_list(self):
        return STORAGE_POOL_DEF

    def fake_api_storage_pool_list(self):
        return STORAGE_POOL_LIST

    def fake_api_resource_dfn_list(self):
        return RESOURCE_DFN_LIST

    def fake_api_snapshot_list(self):
        return SNAPSHOT_LIST_RESP


class LinstorFakeResource(object):

    def __init__(self):
        self.volumes = [{'size': 1069547520}]
        self.id = 0

    def delete(self):
        return True

    def is_diskless(self, host):
        if host in DISKLESS_LIST_RESP:
            return True
        else:
            return False


class LinstorBaseDriverTestCase(test.TestCase):

    def __init__(self, *args, **kwargs):
        super(LinstorBaseDriverTestCase, self).__init__(*args, **kwargs)

    def setUp(self):
        super(LinstorBaseDriverTestCase, self).setUp()

        if drv is None:
            return

        self._mock = mock.Mock()
        self._fake_driver = LinstorAPIFakeDriver()

        self.configuration = mock.Mock(conf.Configuration)

        self.driver = drv.LinstorBaseDriver(
            configuration=self.configuration)
        self.driver.VERSION = '0.0.7'
        self.driver.default_rsc_size = 1
        self.driver.default_vg_name = 'vg-1'
        self.driver.default_downsize_factor = int('4096')
        self.driver.default_pool = STORAGE_POOL_DEF_RESP[0]
        self.driver.host_name = 'node-1'
        self.driver.diskless = True
        self.driver.default_uri = 'linstor://localhost'
        self.driver.default_backend_name = 'lin-test-driver'
        self.driver.configuration.reserved_percentage = 0
        self.driver.configuration.max_over_subscription_ratio = 0
        self.driver.ap_count = 0

    @mock.patch(DRIVER + 'LinstorBaseDriver._ping')
    def test_ping(self, m_ping):
        m_ping.return_value = self._fake_driver.fake_api_ping()

        val = self.driver._ping()
        expected = 1234
        self.assertEqual(expected, val)

    @mock.patch('uuid.uuid4')
    def test_clean_uuid(self, m_uuid):
        m_uuid.return_value = u'bd6472d1-dc3c-4d41-a5f0-f44271c05680'

        val = self.driver._clean_uuid()
        expected = u'bd6472d1-dc3c-4d41-a5f0-f44271c05680'
        self.assertEqual(expected, val)

    @mock.patch('uuid.uuid4')
    def test_clean_uuid_with_braces(self, m_uuid):
        m_uuid.return_value = u'{bd6472d1-dc3c-4d41-a5f0-f44271c05680}'

        val = self.driver._clean_uuid()
        expected = u'bd6472d1-dc3c-4d41-a5f0-f44271c05680'

        m_uuid.assert_called_once()
        self.assertEqual(expected, val)

    # Test volume size conversions
    def test_unit_conversions_to_linstor_1GiB(self):
        val = self.driver._vol_size_to_linstor(1)
        expected = 1044480   # 1048575 - 4096
        self.assertEqual(expected, val)

    def test_unit_conversions_to_linstor_2GiB(self):
        val = self.driver._vol_size_to_linstor(2)
        expected = 2093056   # 2097152 - 4096
        self.assertEqual(expected, val)

    def test_unit_conversions_to_cinder(self):
        val = self.driver._vol_size_to_cinder(1048576)
        expected = 1
        self.assertEqual(expected, val)

    def test_unit_conversions_to_cinder_2GiB(self):
        val = self.driver._vol_size_to_cinder(2097152)
        expected = 2
        self.assertEqual(expected, val)

    def test_is_clean_volume_name(self):
        val = self.driver._is_clean_volume_name(VOLUME_NAMES['cinder'],
                                                drv.DM_VN_PREFIX)
        expected = VOLUME_NAMES['linstor']
        self.assertEqual(expected, val)

    def test_is_clean_volume_name_invalid(self):
        wrong_uuid = 'bc3015e6-695f-4688-91f2-invaliduuid1'
        val = self.driver._is_clean_volume_name(wrong_uuid,
                                                drv.DM_VN_PREFIX)
        expected = None
        self.assertEqual(expected, val)

    def test_snapshot_name_from_cinder_snapshot(self):
        val = self.driver._snapshot_name_from_cinder_snapshot(
            SNAPSHOT)
        expected = VOLUME_NAMES['snap']
        self.assertEqual(expected, val)

    def test_cinder_volume_name_from_drbd_resource(self):
        val = self.driver._cinder_volume_name_from_drbd_resource(
            VOLUME_NAMES['linstor'])
        expected = VOLUME_NAMES['cinder']
        self.assertEqual(expected, val)

    def test_drbd_resource_name_from_cinder_snapshot(self):
        val = self.driver._drbd_resource_name_from_cinder_snapshot(
            SNAPSHOT)
        expected = VOLUME_NAMES['linstor']
        self.assertEqual(expected, val)

    def test_drbd_resource_name_from_cinder_volume(self):
        val = self.driver._drbd_resource_name_from_cinder_volume(
            CINDER_VOLUME)
        expected = VOLUME_NAMES['linstor']
        self.assertEqual(expected, val)

    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_resource_list')
    def test_get_rcs_path(self, m_rsc_list):
        m_rsc_list.return_value = self._fake_driver.fake_api_resource_list()

        val = self.driver._get_rsc_path(VOLUME_NAMES['linstor'])
        expected = '/dev/drbd1005'

        m_rsc_list.assert_called_once()
        self.assertEqual(expected, val)

    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_resource_list')
    def test_get_local_path(self, m_rsc_list):
        m_rsc_list.return_value = self._fake_driver.fake_api_resource_list()

        val = self.driver._get_local_path(CINDER_VOLUME)
        expected = '/dev/drbd1005'

        m_rsc_list.assert_called_once()
        self.assertEqual(expected, val)

    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_storage_pool_dfn_list')
    def test_get_spd(self, m_spd_list):
        m_spd_list.return_value = (
            self._fake_driver.fake_api_storage_pool_dfn_list())

        val = self.driver._get_spd()
        expected = STORAGE_POOL_DEF_RESP

        m_spd_list.assert_called_once()
        self.assertEqual(expected, val)

    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_storage_pool_list')
    def test_get_storage_pool(self, m_sp_list):
        m_sp_list.return_value = (
            self._fake_driver.fake_api_storage_pool_list())

        val = self.driver._get_storage_pool()
        expected = STORAGE_POOL_LIST_RESP

        m_sp_list.assert_called_once()
        self.assertEqual(expected, val)

    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_resource_dfn_list')
    def test_get_resource_definitions(self, m_rscd_list):
        m_rscd_list.return_value = (
            self._fake_driver.fake_api_resource_dfn_list())

        val = self.driver._get_resource_definitions()
        expected = RESOURCE_DFN_LIST_RESP

        m_rscd_list.assert_called_once()
        self.assertEqual(expected, val)

    @mock.patch(DRIVER + 'LinstorBaseDriver._get_snapshot_nodes')
    def test_get_snapshot_nodes(self, m_rsc_list):
        m_rsc_list.return_value = self._fake_driver.fake_api_snapshot_list()

        val = self.driver._get_snapshot_nodes(VOLUME_NAMES['linstor'])
        expected = SNAPSHOT_LIST_RESP

        m_rsc_list.assert_called_once()
        self.assertEqual(expected, val)

    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_resource_list')
    def test_get_diskless_nodes(self, m_rsc_list):
        m_rsc_list.return_value = self._fake_driver.fake_api_resource_list()

        val = self.driver._get_diskless_nodes(RESOURCE['name'])
        expected = DISKLESS_LIST_RESP

        m_rsc_list.assert_called_once()
        self.assertEqual(expected, val)

    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_node_list')
    def test_get_linstor_nodes(self, m_node_list):
        m_node_list.return_value = self._fake_driver.fake_api_node_list()

        val = self.driver._get_linstor_nodes()
        expected = RESOURCE_LIST_RESP

        m_node_list.assert_called_once()
        self.assertEqual(expected, val)

    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_node_list')
    def test_get_nodes(self, m_node_list):
        m_node_list.return_value = self._fake_driver.fake_api_node_list()

        val = self.driver._get_nodes()
        expected = NODES_RESP

        m_node_list.assert_called_once()
        self.assertEqual(expected, val)

    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_size')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_is_diskless')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_resource_list')
    @mock.patch(DRIVER + 'LinstorBaseDriver.get_goodness_function')
    @mock.patch(DRIVER + 'LinstorBaseDriver.get_filter_function')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_resource_dfn_list')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_storage_pool_list')
    def test_get_volume_stats(self,
                              m_sp_list,
                              m_rscd_list,
                              m_filter,
                              m_goodness,
                              m_rsc_list,
                              m_diskless,
                              m_rsc_size):
        m_sp_list.return_value = (
            self._fake_driver.fake_api_storage_pool_list())
        m_rscd_list.return_value = (
            self._fake_driver.fake_api_resource_dfn_list())
        m_filter.return_value = None
        m_goodness.return_value = None
        m_rsc_list.return_value = RESOURCE_LIST
        m_diskless.return_value = True
        m_rsc_size.return_value = 1069547520

        val = self.driver._get_volume_stats()
        expected = VOLUME_STATS_RESP

        m_sp_list.assert_called_once()
        m_rscd_list.assert_called_once()
        self.assertEqual(expected, val)

    @mock.patch(DRIVER + 'LinstorBaseDriver._api_snapshot_create')
    def test_create_snapshot_fail(self,
                                  m_snap_create):
        m_snap_create.return_value = False

        self.assertRaises(cinder_exception.VolumeBackendAPIException,
                          self.driver.create_snapshot, SNAPSHOT)

    @mock.patch(DRIVER + 'LinstorBaseDriver._api_snapshot_create')
    def test_create_snapshot_success(self,
                                     m_snap_create):
        m_snap_create.return_value = True

        # No exception should be raised
        self.assertIsNone(self.driver.create_snapshot(SNAPSHOT))

    @mock.patch(DRIVER + 'LinstorBaseDriver._api_snapshot_delete')
    def test_delete_snapshot_fail(self,
                                  m_snap_delete):
        m_snap_delete.return_value = False

        self.assertRaises(cinder_exception.VolumeBackendAPIException,
                          self.driver.delete_snapshot, SNAPSHOT)

    @mock.patch(DRIVER + 'LinstorBaseDriver._get_snapshot_nodes')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_snapshot_delete')
    def test_delete_snapshot_success(self,
                                     m_snap_delete,
                                     m_snap_nodes):
        m_snap_delete.return_value = True
        m_snap_nodes.return_value = self._fake_driver.fake_api_snapshot_list()

        # No exception should be raised
        self.driver.delete_snapshot(SNAPSHOT)

    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_dfn_delete')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_snapshot_nodes')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_snapshot_delete')
    def test_delete_snapshot_success_cleanup_rd(self,
                                                m_snap_delete,
                                                m_snap_nodes,
                                                m_rd_delete):
        m_snap_delete.return_value = True
        m_snap_nodes.return_value = []
        m_rd_delete.return_value = None

        # No exception should be raised
        self.driver.delete_snapshot(SNAPSHOT)

        # Resource Definition Delete should run once
        m_rd_delete.assert_called_once()

    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_storage_pool_list')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_volume_dfn_set_sp')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_volume_extend')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_create')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_snapshot_resource_restore')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_linstor_nodes')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_snapshot_volume_dfn_restore')
    @mock.patch(DRIVER + 'LinstorBaseDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_dfn_create')
    def test_create_volume_from_snapshot(self,
                                         m_rsc_dfn_create,
                                         m_api_reply,
                                         m_snap_vd_restore,
                                         m_lin_nodes,
                                         m_snap_rsc_restore,
                                         m_rsc_create,
                                         m_vol_extend,
                                         m_vol_dfn,
                                         m_sp_list):
        m_rsc_dfn_create.return_value = True
        m_api_reply.return_value = True
        m_snap_vd_restore.return_value = True
        m_nodes = []
        m_lin_nodes.return_value = m_nodes
        m_snap_rsc_restore.return_value = True
        m_rsc_create.return_value = True
        m_vol_extend.return_value = True
        m_vol_dfn.return_value = True
        m_sp_list.return_value = (
            self._fake_driver.fake_api_storage_pool_list())

        # No exception should be raised
        self.assertIsNone(self.driver.create_volume_from_snapshot(
            CINDER_VOLUME, SNAPSHOT))

    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_storage_pool_list')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_volume_dfn_set_sp')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_volume_extend')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_create')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_snapshot_resource_restore')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_linstor_nodes')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_snapshot_volume_dfn_restore')
    @mock.patch(DRIVER + 'LinstorBaseDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_dfn_create')
    def test_create_volume_from_snapshot_fail_restore(self,
                                                      m_rsc_dfn_create,
                                                      m_api_reply,
                                                      m_snap_vd_restore,
                                                      m_lin_nodes,
                                                      m_snap_rsc_restore,
                                                      m_rsc_create,
                                                      m_vol_extend,
                                                      m_vol_dfn,
                                                      m_sp_list):
        m_rsc_dfn_create.return_value = True
        m_api_reply.return_value = True
        m_snap_vd_restore.return_value = True
        m_nodes = []
        m_lin_nodes.return_value = m_nodes
        m_snap_rsc_restore.return_value = False
        m_rsc_create.return_value = True
        m_vol_extend.return_value = True
        m_vol_dfn.return_value = True
        m_sp_list.return_value = (
            self._fake_driver.fake_api_storage_pool_list())

        # Failing to restore a snapshot should raise an exception
        self.assertRaises(cinder_exception.VolumeBackendAPIException,
                          self.driver.create_volume_from_snapshot,
                          CINDER_VOLUME, SNAPSHOT)

    @mock.patch(DRIVER + 'LinstorBaseDriver.delete_volume')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_storage_pool_list')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_volume_dfn_set_sp')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_volume_extend')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_create')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_snapshot_resource_restore')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_linstor_nodes')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_snapshot_volume_dfn_restore')
    @mock.patch(DRIVER + 'LinstorBaseDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_dfn_create')
    def test_create_volume_from_snapshot_fail_extend(self,
                                                     m_rsc_dfn_create,
                                                     m_api_reply,
                                                     m_snap_vd_restore,
                                                     m_lin_nodes,
                                                     m_snap_rsc_restore,
                                                     m_rsc_create,
                                                     m_vol_extend,
                                                     m_vol_dfn,
                                                     m_sp_list,
                                                     m_delete_volume):
        m_rsc_dfn_create.return_value = True
        m_api_reply.return_value = False
        m_snap_vd_restore.return_value = True
        m_nodes = []
        m_lin_nodes.return_value = m_nodes
        m_snap_rsc_restore.return_value = True
        m_rsc_create.return_value = True
        m_vol_extend.return_value = True
        m_vol_dfn.return_value = True
        m_sp_list.return_value = (
            self._fake_driver.fake_api_storage_pool_list())
        m_delete_volume.return_value = True

        # Failing to extend the volume after a snapshot restoration should
        # raise an exception
        new_volume = CINDER_VOLUME
        new_volume['size'] = 2
        self.assertRaises(cinder_exception.VolumeBackendAPIException,
                          self.driver.create_volume_from_snapshot,
                          new_volume, SNAPSHOT)

    @mock.patch(DRIVER + 'LinstorBaseDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_create')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_volume_dfn_create')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_dfn_create')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_storage_pool_create')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_storage_pool_dfn_list')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_node_list')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_storage_pool_list')
    def test_create_volume_fail_no_linstor_nodes(self,
                                                 m_sp_list,
                                                 m_node_list,
                                                 m_spd_list,
                                                 m_sp_create,
                                                 m_rsc_dfn_create,
                                                 m_vol_dfn_create,
                                                 m_rsc_create,
                                                 m_api_reply):
        m_sp_list.return_value = []
        m_node_list.return_value = []
        m_spd_list.return_value = (
            self._fake_driver.fake_api_storage_pool_dfn_list())
        m_sp_create.return_value = True
        m_rsc_dfn_create.return_value = True
        m_vol_dfn_create.return_value = True
        m_rsc_create.return_value = True
        m_api_reply.return_value = True

        test_volume = CINDER_VOLUME
        test_volume['migration_status'] = ('migrating:',
                                           str(VOLUME_NAMES['cinder']))
        test_volume['display_name'] = 'test_volume'
        test_volume['host'] = 'node_one'
        test_volume['size'] = 1

        self.assertRaises(cinder_exception.VolumeBackendAPIException,
                          self.driver.create_volume, test_volume)

    @mock.patch(DRIVER + 'LinstorBaseDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_create')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_volume_dfn_create')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_dfn_create')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_storage_pool_create')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_storage_pool_dfn_list')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_node_list')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_storage_pool_list')
    def test_create_volume_fail_rsc_create(self,
                                           m_sp_list,
                                           m_node_list,
                                           m_spd_list,
                                           m_sp_create,
                                           m_rsc_dfn_create,
                                           m_vol_dfn_create,
                                           m_rsc_create,
                                           m_api_reply):
        m_sp_list.return_value = (
            self._fake_driver.fake_api_storage_pool_list())
        m_node_list.return_value = self._fake_driver.fake_api_node_list()
        m_spd_list.return_value = (
            self._fake_driver.fake_api_storage_pool_dfn_list())
        m_sp_create.return_value = True
        m_rsc_dfn_create.return_value = True
        m_vol_dfn_create.return_value = True
        m_rsc_create.return_value = True
        m_api_reply.return_value = False

        test_volume = CINDER_VOLUME
        test_volume['migration_status'] = ('migrating:',
                                           str(VOLUME_NAMES['cinder']))
        test_volume['display_name'] = 'test_volume'
        test_volume['host'] = 'node_one'
        test_volume['size'] = 1

        self.assertRaises(cinder_exception.VolumeBackendAPIException,
                          self.driver.create_volume, test_volume)

    @mock.patch(DRIVER + 'LinstorBaseDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_create')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_volume_dfn_create')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_dfn_create')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_storage_pool_create')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_storage_pool_dfn_list')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_node_list')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_storage_pool_list')
    def test_create_volume(self,
                           m_sp_list,
                           m_node_list,
                           m_spd_list,
                           m_sp_create,
                           m_rsc_dfn_create,
                           m_vol_dfn_create,
                           m_rsc_create,
                           m_api_reply):
        m_sp_list.return_value = (
            self._fake_driver.fake_api_storage_pool_list())
        m_node_list.return_value = self._fake_driver.fake_api_node_list()
        m_spd_list.return_value = (
            self._fake_driver.fake_api_storage_pool_dfn_list())
        m_sp_create.return_value = True
        m_rsc_dfn_create.return_value = True
        m_vol_dfn_create.return_value = True
        m_rsc_create.return_value = True
        m_api_reply.return_value = True

        test_volume = CINDER_VOLUME
        test_volume['migration_status'] = ('migrating:',
                                           str(VOLUME_NAMES['cinder']))
        test_volume['display_name'] = 'test_volume'
        test_volume['host'] = 'node_one'
        test_volume['size'] = 1

        val = self.driver.create_volume(test_volume)
        expected = {}
        self.assertEqual(expected, val)

    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_auto_delete')
    @mock.patch(DRIVER + 'LinstorBaseDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_dfn_delete')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_volume_dfn_delete')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_delete')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_resource_list')
    def test_delete_volume_fail_incomplete(self,
                                           m_rsc_list,
                                           m_rsc_delete,
                                           m_vol_dfn_delete,
                                           m_rsc_dfn_delete,
                                           m_api_reply,
                                           m_rsc_auto_delete):
        m_rsc_list.return_value = self._fake_driver.fake_api_resource_list()
        m_rsc_delete.return_value = True
        m_vol_dfn_delete.return_value = True
        m_rsc_dfn_delete.return_value = True
        m_api_reply.return_value = False
        m_rsc_auto_delete.return_value = True

        test_volume = CINDER_VOLUME
        test_volume['display_name'] = 'linstor_test'
        test_volume['host'] = 'node_one'
        test_volume['size'] = 1

        self.assertRaises(cinder_exception.VolumeBackendAPIException,
                          self.driver.delete_volume, test_volume)

    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_auto_delete')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_diskless_nodes')
    @mock.patch(DRIVER + 'LinstorBaseDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_dfn_delete')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_volume_dfn_delete')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_delete')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_resource_list')
    def test_delete_volume_fail_diskless_remove(self,
                                                m_rsc_list,
                                                m_rsc_delete,
                                                m_vol_dfn_delete,
                                                m_rsc_dfn_delete,
                                                m_api_reply,
                                                m_diskless,
                                                m_rsc_auto_delete):
        m_rsc_list.return_value = self._fake_driver.fake_api_resource_list()
        m_rsc_delete.return_value = False
        m_vol_dfn_delete.return_value = True
        m_rsc_dfn_delete.return_value = True
        m_api_reply.return_value = False
        m_diskless.return_value = ['foo']
        m_rsc_auto_delete.return_value = True

        test_volume = CINDER_VOLUME
        test_volume['display_name'] = 'linstor_test'
        test_volume['host'] = 'node_one'
        test_volume['size'] = 1

        # Raises exception for failing to delete a diskless resource
        self.assertRaises(cinder_exception.VolumeBackendAPIException,
                          self.driver.delete_volume, test_volume)

    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_auto_delete')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_snapshot_nodes')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_diskless_nodes')
    @mock.patch(DRIVER + 'LinstorBaseDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_dfn_delete')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_volume_dfn_delete')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_delete')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_resource_list')
    def test_delete_volume_fail_diskful_remove(self,
                                               m_rsc_list,
                                               m_rsc_delete,
                                               m_vol_dfn_delete,
                                               m_rsc_dfn_delete,
                                               m_api_reply,
                                               m_diskless,
                                               m_snap_nodes,
                                               m_rsc_auto_delete):
        m_rsc_list.return_value = self._fake_driver.fake_api_resource_list()
        m_rsc_delete.return_value = False
        m_vol_dfn_delete.return_value = True
        m_rsc_dfn_delete.return_value = True
        m_api_reply.return_value = False
        m_diskless.return_value = []
        m_snap_nodes.return_value = ['foo']
        m_rsc_auto_delete.return_value = True

        test_volume = CINDER_VOLUME
        test_volume['display_name'] = 'linstor_test'
        test_volume['host'] = 'node_one'
        test_volume['size'] = 1

        # Raises exception for failing to delete a diskful resource
        self.assertRaises(cinder_exception.VolumeBackendAPIException,
                          self.driver.delete_volume, test_volume)

    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_auto_delete')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_snapshot_nodes')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_diskless_nodes')
    @mock.patch(DRIVER + 'LinstorBaseDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_dfn_delete')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_volume_dfn_delete')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_delete')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_resource_list')
    def test_delete_volume_fail_volume_definition(self,
                                                  m_rsc_list,
                                                  m_rsc_delete,
                                                  m_vol_dfn_delete,
                                                  m_rsc_dfn_delete,
                                                  m_api_reply,
                                                  m_diskless,
                                                  m_snap_nodes,
                                                  m_rsc_auto_delete):
        m_rsc_list.return_value = self._fake_driver.fake_api_resource_list()
        m_rsc_delete.return_value = True
        m_vol_dfn_delete.return_value = False
        m_rsc_dfn_delete.return_value = True
        m_api_reply.return_value = False
        m_diskless.return_value = []
        m_snap_nodes.return_value = []
        m_rsc_auto_delete.return_value = True

        test_volume = CINDER_VOLUME
        test_volume['display_name'] = 'linstor_test'
        test_volume['host'] = 'node_one'
        test_volume['size'] = 1

        # Raises exception for failing to delete a volume definition
        self.assertRaises(cinder_exception.VolumeBackendAPIException,
                          self.driver.delete_volume, test_volume)

    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_auto_delete')
    @mock.patch(DRIVER + 'LinstorBaseDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_dfn_delete')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_volume_dfn_delete')
    @mock.patch(DRIVER + 'LinstorBaseDriver._api_rsc_delete')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_resource_list')
    def test_delete_volume(self,
                           m_rsc_list,
                           m_rsc_delete,
                           m_vol_dfn_delete,
                           m_rsc_dfn_delete,
                           m_api_reply,
                           m_rsc_auto_delete):
        m_rsc_list.return_value = self._fake_driver.fake_api_resource_list()
        m_rsc_delete.return_value = True
        m_vol_dfn_delete.return_value = True
        m_rsc_dfn_delete.return_value = True
        m_api_reply.return_value = True
        m_rsc_auto_delete.return_value = True

        test_volume = CINDER_VOLUME
        test_volume['display_name'] = 'linstor_test'
        test_volume['host'] = 'node_one'
        test_volume['size'] = 1

        val = self.driver.delete_volume(test_volume)
        expected = True
        self.assertEqual(expected, val)

    @mock.patch(DRIVER + 'LinstorBaseDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_volume_extend')
    def test_extend_volume_success(self, m_vol_extend, m_api_reply):
        m_vol_extend.return_value = True
        m_api_reply.return_value = True

        # No exception should be raised
        self.driver.extend_volume(CINDER_VOLUME, 2)

    @mock.patch(DRIVER + 'LinstorBaseDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorBaseDriver._get_api_volume_extend')
    def test_extend_volume_fail(self, m_vol_extend, m_api_reply):
        m_vol_extend.return_value = False
        m_api_reply.return_value = False

        self.assertRaises(cinder_exception.VolumeBackendAPIException,
                          self.driver.extend_volume,
                          CINDER_VOLUME,
                          2)

    def test_migrate_volume(self):
        m_ctxt = {}
        m_volume = {}
        m_host = ''

        val = self.driver.migrate_volume(m_ctxt, m_volume, m_host)
        expected = (False, None)
        self.assertEqual(expected, val)


class LinstorIscsiDriverTestCase(test.TestCase):

    def __init__(self, *args, **kwargs):
        super(LinstorIscsiDriverTestCase, self).__init__(*args, **kwargs)

    def setUp(self):
        super(LinstorIscsiDriverTestCase, self).setUp()

        self._mock = mock.Mock()
        self._fake_driver = LinstorAPIFakeDriver()

        self.configuration = mock.Mock(conf.Configuration)
        self.configuration.iscsi_helper = 'tgtadm'
        self.driver = drv.LinstorIscsiDriver(
            configuration=self.configuration, h_name='tgtadm')

        self.driver.VERSION = '0.0.7'
        self.driver.default_rsc_size = 1
        self.driver.default_vg_name = 'vg-1'
        self.driver.default_downsize_factor = int('4096')
        self.driver.default_pool = STORAGE_POOL_DEF_RESP[0]
        self.driver.host_name = 'node_one'
        self.driver.diskless = True
        self.driver.location_info = 'LinstorIscsi:linstor://localhost'
        self.driver.default_backend_name = 'lin-test-driver'
        self.driver.configuration.reserved_percentage = int('0')
        self.driver.configuration.max_over_subscription_ratio = int('0')

    @mock.patch(DRIVER + 'LinstorIscsiDriver._get_api_resource_list')
    @mock.patch(DRIVER + 'LinstorIscsiDriver._get_volume_stats')
    def test_iscsi_get_volume_stats(self, m_vol_stats, m_rsc_list):

        m_vol_stats.return_value = VOLUME_STATS_RESP
        m_rsc_list.return_value = RESOURCE_LIST

        val = self.driver.get_volume_stats()

        expected = VOLUME_STATS_RESP
        expected["storage_protocol"] = 'iSCSI'
        self.assertEqual(expected, val)

    @mock.patch(DRIVER + 'linstor')
    def test_iscsi_check_for_setup_error_pass(self, m_linstor):
        m_linstor.return_value = True

        # No exception should be raised
        self.driver.check_for_setup_error()


class LinstorDrbdDriverTestCase(test.TestCase):

    def __init__(self, *args, **kwargs):
        super(LinstorDrbdDriverTestCase, self).__init__(*args, **kwargs)

    def setUp(self):
        super(LinstorDrbdDriverTestCase, self).setUp()

        self._mock = mock.Mock()
        self._fake_driver = LinstorAPIFakeDriver()

        self.configuration = mock.Mock(conf.Configuration)
        self.driver = drv.LinstorDrbdDriver(
            configuration=self.configuration)

        self.driver.VERSION = '0.0.7'
        self.driver.default_rsc_size = 1
        self.driver.default_vg_name = 'vg-1'
        self.driver.default_downsize_factor = int('4096')
        self.driver.default_pool = STORAGE_POOL_DEF_RESP[0]
        self.driver.host_name = 'node_one'
        self.driver.diskless = True
        self.driver.location_info = 'LinstorDrbd:linstor://localhost'
        self.driver.default_backend_name = 'lin-test-driver'
        self.driver.configuration.reserved_percentage = int('0')
        self.driver.configuration.max_over_subscription_ratio = int('0')

    @mock.patch(DRIVER + 'LinstorDrbdDriver._get_rsc_path')
    def test_drbd_return_drbd_config(self, m_rsc_path):
        m_rsc_path.return_value = '/dev/drbd1005'

        val = self.driver._return_drbd_config(CINDER_VOLUME)

        expected = {
            'driver_volume_type': 'local',
            'data': {
                "device_path": str(m_rsc_path.return_value)
            }
        }
        self.assertEqual(expected, val)

    @mock.patch(DRIVER + 'LinstorDrbdDriver._get_api_storage_pool_list')
    def test_drbd_node_in_sp(self, m_sp_list):
        m_sp_list.return_value = (
            self._fake_driver.fake_api_storage_pool_list())

        val = self.driver._node_in_sp('node-1')
        self.assertTrue(val)

    @mock.patch(DRIVER + 'LinstorDrbdDriver._get_volume_stats')
    def test_drbd_get_volume_stats(self, m_vol_stats):
        m_vol_stats.return_value = VOLUME_STATS_RESP

        val = self.driver.get_volume_stats()
        expected = VOLUME_STATS_RESP
        expected["storage_protocol"] = 'DRBD'
        self.assertEqual(expected, val)

    @mock.patch(DRIVER + 'linstor')
    def test_drbd_check_for_setup_error_pass(self, m_linstor):
        m_linstor.return_value = True

        # No exception should be raised
        self.driver.check_for_setup_error()

    @mock.patch(DRIVER + 'LinstorDrbdDriver._get_rsc_path')
    @mock.patch(DRIVER + 'LinstorDrbdDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorDrbdDriver._api_rsc_create')
    @mock.patch(DRIVER + 'LinstorDrbdDriver._node_in_sp')
    def test_drbd_initialize_connection_pass(self,
                                             m_node_sp,
                                             m_rsc_create,
                                             m_check,
                                             m_rsc_path):
        m_node_sp.return_value = True
        m_rsc_create.return_value = True
        m_check.return_value = True
        m_rsc_path.return_value = '/dev/drbd1000'

        connector = {}
        connector["host"] = 'wp-u16-cinder-dev-lg'

        val = self.driver.initialize_connection(CINDER_VOLUME, connector)

        expected = {
            'driver_volume_type': 'local',
            'data': {
                "device_path": str(m_rsc_path.return_value)
            }
        }
        self.assertEqual(expected, val)

    @mock.patch(DRIVER + 'LinstorDrbdDriver._check_api_reply')
    @mock.patch(DRIVER + 'LinstorDrbdDriver._api_rsc_delete')
    @mock.patch(DRIVER + 'LinstorDrbdDriver._node_in_sp')
    def test_drbd_terminate_connection_pass(self,
                                            m_node_sp,
                                            m_rsc_create,
                                            m_check):
        m_node_sp.return_value = True
        m_rsc_create.return_value = True
        m_check.return_value = True

        connector = {}
        connector["host"] = 'wp-u16-cinder-dev-lg'

        # No exception should be raised
        self.driver.terminate_connection(CINDER_VOLUME, connector)



# Copyright (C) 2016 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinder.api import common
from cinder.api import microversions as mv
from cinder.policies import groups as policy
from cinder import utils


class ViewBuilder(common.ViewBuilder):
    """Model group API responses as a python dictionary."""

    _collection_name = "groups"

    def __init__(self):
        """Initialize view builder."""
        super(ViewBuilder, self).__init__()

    def summary_list(self, request, groups):
        """Show a list of groups without many details."""
        return self._list_view(self.summary, request, groups)

    def detail_list(self, request, groups):
        """Detailed view of a list of groups ."""
        return self._list_view(self.detail, request, groups)

    def summary(self, request, group):
        """Generic, non-detailed view of a group."""
        return {
            'group': {
                'id': group.id,
                'name': group.name
            }
        }

    def detail(self, request, group):
        """Detailed view of a single group."""
        context = request.environ['cinder.context']
        group_ref = {
            'group': {
                'id': group.id,
                'status': group.status,
                'availability_zone': group.availability_zone,
                'created_at': group.created_at,
                'name': group.name,
                'description': group.description,
                'group_type': group.group_type_id,
                'volume_types': [v_type.id for v_type in group.volume_types],
            }
        }

        req_version = request.api_version_request
        # Add group_snapshot_id and source_group_id if min version is greater
        # than or equal to GROUP_SNAPSHOTS.
        if req_version.matches(mv.GROUP_SNAPSHOTS, None):
            group_ref['group']['group_snapshot_id'] = group.group_snapshot_id
            group_ref['group']['source_group_id'] = group.source_group_id

        # Add volumes if min version is greater than or equal to
        # GROUP_VOLUME_LIST.
        if req_version.matches(mv.GROUP_VOLUME_LIST, None):
            if utils.get_bool_param('list_volume', request.params):
                group_ref['group']['volumes'] = [volume.id
                                                 for volume in group.volumes]

        # Add replication_status if min version is greater than or equal
        # to GROUP_REPLICATION.
        if req_version.matches(mv.GROUP_REPLICATION, None):
            group_ref['group']['replication_status'] = group.replication_status

        if req_version.matches(mv.GROUP_GROUPSNAPSHOT_PROJECT_ID, None):
            if context.authorize(policy.GROUP_ATTRIBUTES_POLICY, fatal=False):
                group_ref['group']['project_id'] = group.project_id

        return group_ref

    def _list_view(self, func, request, groups):
        """Provide a view for a list of groups."""
        groups_list = [
            func(request, group)['group']
            for group in groups]
        grp_links = self._get_collection_links(request,
                                               groups,
                                               self._collection_name)
        groups_dict = dict(groups=groups_list)
        if grp_links:
            groups_dict['group_links'] = grp_links

        return groups_dict



#   Copyright 2014 IBM Corp.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from oslo_log import log as logging
from six.moves import http_client
import webob

from cinder.api import extensions
from cinder.api.openstack import wsgi
from cinder.policies import manageable_volumes as policy
from cinder import volume

LOG = logging.getLogger(__name__)


class VolumeUnmanageController(wsgi.Controller):
    def __init__(self, *args, **kwargs):
        super(VolumeUnmanageController, self).__init__(*args, **kwargs)
        self.volume_api = volume.API()

    @wsgi.response(http_client.ACCEPTED)
    @wsgi.action('os-unmanage')
    def unmanage(self, req, id, body):
        """Stop managing a volume.

        This action is very much like a delete, except that a different
        method (unmanage) is called on the Cinder driver.  This has the effect
        of removing the volume from Cinder management without actually
        removing the backend storage object associated with it.

        There are no required parameters.

        A Not Found error is returned if the specified volume does not exist.

        A Bad Request error is returned if the specified volume is still
        attached to an instance.
        """
        context = req.environ['cinder.context']

        LOG.info("Unmanage volume with id: %s", id)

        # Not found exception will be handled at the wsgi level
        vol = self.volume_api.get(context, id)
        context.authorize(policy.UNMANAGE_POLICY, target_obj=vol)
        self.volume_api.delete(context, vol, unmanage_only=True)
        return webob.Response(status_int=http_client.ACCEPTED)


class Volume_unmanage(extensions.ExtensionDescriptor):
    """Enable volume unmanage operation."""

    name = "VolumeUnmanage"
    alias = "os-volume-unmanage"
    updated = "2012-05-31T00:00:00+00:00"

    def get_controller_extensions(self):
        controller = VolumeUnmanageController()
        extension = extensions.ControllerExtension(self, 'volumes', controller)
        return [extension]



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class ViewBuilder(object):
    """Model an resource filters API response as a python dictionary."""

    _collection_name = "resource_filters"

    @classmethod
    def list(cls, filters):
        """Build a view of a list of resource filters.

         .. code-block:: json

            {
               "resource_filters": [{
                   "resource": "resource_1",
                   "filters": ["filter1", "filter2", "filter3"]
                }]
            }
        """

        return {'resource_filters': [{
            'resource': fil[0],
            'filters': fil[1]} for fil in filters.items()]}
